import copy
import json
import logging
from collections import defaultdict
from pathlib import Path

import rich
from jinja2 import FileSystemLoader, PrefixLoader
from rich.progress import Progress

from .collection import Collection
from .data_object import DataObject
from .engine import engine
from .page import Page
from .plugins import PluginManager, handle_plugin_registration
from .site_map import SiteMap
from .themes import Theme, ThemeManager

try:
    # Get the RE version for display. If it's not set it means we're working locally.
    from render_engine.__version__ import __version__ as re_version
except ImportError:
    re_version = "development"


class Site:
    """
    The site stores your pages and collections to be rendered.

    Attributes:
        site_vars (dict): A dictionary containing site-wide variables and their values.
        plugin_settings (dict): A dictionary containing plugin settings.
        render_html_site_map (bool): Whether to render the generated site map as an HTML page.
        render_xml_site_map (bool): Whether to render the generated site map as XML.

    Methods:
        update_site_vars(**kwargs): Updates the site-wide variables with the given key-value pairs.
        register_plugins(*plugins, **plugin_settings): Registers the specified plugins with the site.
        register_theme(theme): Registers a theme with the site.
        register_themes(*themes): Registers multiple themes with the site.
        update_theme_settings(**settings): Updates the theme settings with the given key-value pairs.
        collection(Collection): Adds a collection to the site's route list.
        page(Page): Adds a page to the site's route list.
        load_themes(): Loads the themes registered with the site.
        render(): Renders all pages and collections added to the site.

    Properties:
        output_path: The output path where the rendered files will be saved.
        static_paths: The paths to static files used in the site.
        template_path: The path to the template files used for rendering.
    """

    site_vars: dict = {
        "SITE_TITLE": "Untitled Site",
        "SITE_URL": "http://localhost:8000/",
        "DATETIME_FORMAT": "%d %b %Y %H:%M %Z",
        "head": set(),
        "theme": {},
    }
    _output_path: str | Path = "output"
    _template_path: str | Path = "templates"
    _static_paths: set = {"static"}
    plugin_settings: dict = {"plugins": defaultdict(dict)}
    render_html_site_map: bool = False
    render_xml_site_map: bool = False

    def __init__(
        self,
    ) -> None:
        self.plugin_manager: PluginManager = PluginManager()
        self.theme_manager = ThemeManager(
            engine=engine,
            output_path=self._output_path,
            static_paths=self._static_paths,
        )
        self.route_list: dict = {}
        self.site_settings: dict = {}
        self.subcollections: dict[str, list] = {"pages": []}
        self.theme_manager.engine.globals.update(self.site_vars)
        self.theme_manager.add_loader(0, FileSystemLoader(self._template_path))
        self._site_map = SiteMap()

    @property
    def output_path(self) -> Path | str:
        return self.theme_manager.output_path

    @output_path.setter
    def output_path(self, output_path: Path | str) -> None:
        self.theme_manager.output_path = output_path

    @property
    def static_paths(self) -> set:
        return self.theme_manager.static_paths

    @static_paths.setter
    def static_paths(self, static_paths: set) -> None:
        self.theme_manager.static_paths = static_paths

    @property
    def site_map(self) -> SiteMap | None:
        return self._site_map

    def update_site_vars(self, **kwargs) -> None:
        self.site_vars.update(**kwargs)
        self.theme_manager.engine.globals.update(self.site_vars)

    def register_plugins(self, *plugins, **plugin_settings) -> None:
        """
        Register plugins for the site

        :param plugins: List of plugins to register
        :param plugin_settings: KW arguments where the key is the plugin name and the value is a dictionary of settings.
        """
        handle_plugin_registration(
            self.plugin_manager,
            [(plugin, plugin_settings.get(plugin.__name__, dict())) for plugin in plugins],
            self.plugin_settings,
        )

    def register_theme(self, theme: Theme) -> None:
        """Overrides the ThemeManager register_theme method to add plugins to the site"""
        self.theme_manager.register_theme(theme)

        if theme.plugins:
            self.register_plugins(*theme.plugins)

    def register_themes(self, *themes: Theme) -> None:
        """
        Register multiple themes.

        Args:
            *themes: Theme objects to register
        """
        for theme in themes:
            self.register_theme(theme)

    def update_theme_settings(self, **settings) -> None:
        for key, value in settings.items():
            self.site_vars["theme"].update({key: value})

    def collection(self, Collection: type[Collection]) -> Collection:
        """
        Add the collection to the route list to be rendered later.

        This is the primary way to add a collection to the site and
        can either be called on an uninstantiated class or on the class definition as a decorator.

        In most cases. You should use the decorator method.

        ```python
        from render_engine import Site, Collection

        site = Site()

        @site.collection # works
        class Pages(Collection):
            pass


        class Posts(Collection):
            pass

        site.collection(Posts) # also works
        ```
        """
        _Collection = Collection()
        _Collection.plugin_manager = copy.deepcopy(self.plugin_manager)
        self.register_themes(*getattr(_Collection, "required_themes", []))

        if plugins := getattr(_Collection, "plugins", []):
            handle_plugin_registration(
                _Collection.plugin_manager,
                plugins,
                getattr(_Collection, "plugin_settings", dict()),
            )

        for plugin in getattr(_Collection, "ignore_plugins", []):
            _Collection.plugin_manager.unregister_plugin(plugin)

        self.route_list[_Collection._slug] = _Collection
        return _Collection

    def page(self, _page: type[Page]) -> Page:
        """
        Add the page to the route list to be rendered later.
        Also remaps `title` in case the user wants to use it in the template rendering.

        This is the primary way to add a page to the site and can either be called
        on an uninstantiated class or on the class definition as a decorator.

        In most cases. You should use the decorator method.

        ```python

        from render_engine import Site, Page

        site = Site()

        @site.page # works
        class Home(Page):
            pass

        class About(Page):
            pass

        site.page(About) # also works
        ```
        """
        page = _page()
        page.title = page._title  # Expose _title to the user through `title`

        # copy the plugin manager, removing any plugins that the page has ignored
        page.plugin_manager = copy.deepcopy(self.plugin_manager)

        if plugins := getattr(page, "plugins", []):
            handle_plugin_registration(page.plugin_manager, plugins, getattr(page, "plugin_settings", dict()))

        for plugin in getattr(page, "ignore_plugins", []):
            page.plugin_manager.unregister_plugin(plugin)

        self.route_list[getattr(page, page._reference)] = page
        return page

    def data_object(self, _data_object: type[DataObject]) -> DataObject:
        """
        Add the data object to the route list to be rendered later.

        This is the primary way to add a data object to the site and can either be called
        on an uninstantiated class or on the class definition as a decorator.

        In most cases. You should use the decorator method.

        ```python

        from render_engine import Site, DataObject

        site = Site()

        data_object = {'key': 'value'}

        @site.data_object  # works
        class MyJson(DataObject):
            data_object = data_object

        class MyJson(DataObject):
            data_object = data_object

        site.page(MyJson) # also works
        ```
        """
        # Since adding a function as a raw attribute on a class will make it a method we need to actually handle
        # it with `staticmethod`. Rather than asking users to do that, we will extract whatever might be set as the
        # serializer function and use the `__init__` for the `DataObject` to set it so that it is properly handled.
        serializer = getattr(_data_object, "serializer", None) or json.dumps
        data_object = _data_object(serializer)

        data_object.plugin_manager = copy.deepcopy(self.plugin_manager)
        if plugins := getattr(data_object, "plugins", []):
            handle_plugin_registration(
                data_object.plugin_manager, plugins, getattr(data_object, "plugin_settings", dict())
            )

        for plugin in getattr(data_object, "ignore_plugins", []):
            data_object.plugin_manager.unregister_plugin(plugin)

        self.route_list[data_object.path_name] = data_object
        return data_object

    def load_themes(self) -> None:
        """
        function for registering the themes with the theme_manager.
        Used prior to rendering and cli-tasks
        """
        # load themes in the ChoiceLoader/FileLoader
        for theme_prefix, theme_loader in self.theme_manager.prefix.items():
            logging.info(f"loading theme: {theme_prefix}")
            self.theme_manager.add_loader(-1, theme_loader)
        # load themes in the PrefixLoader
        self.theme_manager.add_loader(-1, PrefixLoader(self.theme_manager.prefix))

    @property
    def template_path(self) -> str:
        return self.theme_manager.template_path

    @template_path.setter
    def template_path(self, template_path: str) -> None:
        self.theme_manager.add_loader(0, FileSystemLoader(template_path))

    def render(self, site_url: str | None = None) -> None:
        """
        Render all pages and collections.

        These are pages and collections that have been added to the site using
        the [`Site.page`][src.render_engine.Site.page]
        and [`Site.collection`][src.render_engine.Site.collection] methods.

        Render should be called after all pages and collections have been added to the site.

        You can choose to call it manually in your file or
        use the CLI command [`render-engine build`][src.render_engine.cli.build]

        :param site_url: Alternate URL for the site to use in the site map
        """
        rich.print(
            f"[green]Building {repr(self.site_vars.get('SITE_TITLE', 'your site'))} "
            f"with Render Engine version {re_version}"
        )
        with Progress() as progress:
            site_url = site_url if site_url is not None else self.site_vars.get("SITE_URL", "")
            task_site_map = progress.add_task(f"Updating site map. {site_url=}", total=1)

            # self._site_map will be initialized with an empty route list and the site URL pointing
            # to https://localhost:8000/ This task will update to the correct site URL and with the route list
            # as it will be rendered.
            self._site_map.site_url = site_url
            self._site_map.update(self.route_list)

            if self.render_html_site_map:

                @self.page
                class SiteMapPage(Page):
                    title = f"{self.site_vars.get('SITE_TITLE', '')} Site Map"
                    path_name = "site_map.html"
                    content = self._site_map.html
                    template = "page.html"

            if self.render_xml_site_map:

                @self.page
                class SiteMapXml(Page):
                    path_name = "site_map.xml"
                    template = "sitemap.xml"

            progress.update(task_site_map, advance=1)

            pre_build_task = progress.add_task("Loading Pre-Build Plugins and Themes", total=1)
            self.plugin_manager.hook.pre_build_site(
                site=self,
                settings=self.plugin_manager.plugin_settings,
            )

            self.load_themes()
            self.theme_manager.engine.globals.update(self.site_vars)
            progress.update(pre_build_task, advance=1)
            # Parse Route List
            task_add_route = progress.add_task("[blue]Adding Routes", total=len(self.route_list))

            self.theme_manager._render_static()

            self.theme_manager.engine.globals["site"] = self
            self.theme_manager.engine.globals["routes"] = self.route_list

            for slug, entry in self.route_list.items():
                entry.site = self
                progress.update(task_add_route, description=f"[blue]Adding[gold]Route: [blue]{slug}")
                args = []
                match entry:
                    case Page():
                        progress.update(
                            task_add_route,
                            description=f"[blue]Adding[gold]Route: [blue]{entry._slug}",
                        )
                        args = [self.theme_manager]
                    case Collection():
                        progress.update(
                            task_add_route,
                            description=f"[blue]Adding[gold]Route: [blue]Collection {entry._slug}",
                        )
                        pre_build_collection_task = progress.add_task(
                            "Loading Pre-Build-Collection Plugins",
                            total=1,
                        )
                        entry._run_collection_plugins(
                            hook_type="pre_build_collection",
                            site=self,
                        )
                        progress.update(pre_build_collection_task, advance=1)
                    case DataObject():
                        progress.update(
                            task_add_route,
                            description=f"[blue]Adding[gold]Route: [blue]{entry.filename}",
                        )

                entry.render(*args)
                if isinstance(entry, Collection):
                    post_build_collection_task = progress.add_task(
                        "Loading Post-Build-Collection Plugins",
                        total=1,
                    )
                    entry._run_collection_plugins(
                        hook_type="post_build_collection",
                        site=self,
                    )
                    progress.update(post_build_collection_task, advance=1)
                progress.update(task_add_route, advance=1)

            post_build_task = progress.add_task("Loading Post-Build Plugins", total=1)
            self.plugin_manager.hook.post_build_site(
                site=self,
                settings=self.plugin_manager.plugin_settings,
            )
            progress.update(post_build_task, advance=1)



import datetime

import pytest

from render_engine.blog import Blog
from render_engine.page import Page


@pytest.fixture()
def blog_with_pages():
    class Page1(Page):
        title = "Older Blog Post"
        date = datetime.datetime(2024, 1, 1)
        content = """Older Page"""

    class Page2(Page):
        title = "Newer Blog Post"
        date = datetime.datetime(2024, 1, 2)
        content = """Newer Page"""

    class CustomBlog(Blog):
        pages = [Page1(), Page2()]

    return CustomBlog()


def test_blog_has_feed():
    """Tests that the Blog class has a Feed attribute."""

    blog = Blog()
    assert hasattr(blog, "Feed")


def test_sorting_blog_posts_with_no_date_raises_error():
    """
    Tests that blog posts that don't have date attributes
    raise an error when trying to sort
    """

    class Page1(Page):
        title = "Older Blog Post"
        content = """Older Page"""

    class Page2(Page):
        title = "Newer Blog Post"
        content = """Newer Page"""

    class CustomBlog(Blog):
        pages = [Page1, Page2]

    blog = CustomBlog()

    with pytest.raises(AttributeError):
        print(blog.sorted_pages)


def test_blog_sorted_pages_is_in_reverse(blog_with_pages):
    assert [page.title for page in blog_with_pages.sorted_pages] == [
        "Newer Blog Post",
        "Older Blog Post",
    ]


def test_blog_feed_is_sorted_in_reverse(blog_with_pages):
    """
    Newest Blog Posts should be listed first to
    conform with many services that don't retrieve all the posts
    """

    assert blog_with_pages.feed.pages[0].title == "Newer Blog Post"


def test_blog_works_with_datetime_dates():
    class Page1(Page):
        date = datetime.date(2025, 1, 1)
        title = "Older Blog Post"
        content = """Older Page"""

    class Page2(Page):
        date = datetime.date(2025, 1, 2)
        title = "Newer Blog Post"
        content = """Newer Page"""

    class CustomBlog(Blog):
        pages = [Page1, Page2]

    blog = CustomBlog()

    assert [page.title for page in blog.sorted_pages] == [
        "Newer Blog Post",
        "Older Blog Post",
    ]



from collections import defaultdict

import pytest

from render_engine.collection import Collection
from render_engine.page import Page
from render_engine.site import Site
from render_engine.site_map import SiteMap, SiteMapEntry

PAGE_TEMPLATE = """
---
title: {title}
slug: {slug}
---
{title}
{url}
"""


@pytest.fixture(scope="module")
def site(tmp_path_factory):
    base_temp_path = tmp_path_factory.getbasetemp()
    tmp_output_path = base_temp_path / "test_output"
    tmp_input_path = base_temp_path / "content"
    tmp_input_path.mkdir(parents=True)
    tmp_template_path = base_temp_path / "templates"
    tmp_template_path.mkdir(parents=True)
    template_file_name = "template_file.html"
    template_file = tmp_template_path / template_file_name
    template_file.write_text("{{ content }}\n")

    coll_pages = defaultdict(list)
    for collection in ["coll1", "coll2"]:
        for n in range(4):
            coll_pages[collection].append(
                Page(
                    content=PAGE_TEMPLATE.format(
                        title=f"{collection} -- Page {n}",
                        slug=f"page{n}",
                        url="{{ site_map.find('page0', collection='coll1').url_for }}",
                    )
                )
            )
    for n in range(4):
        content_file = tmp_input_path / f"page{n}.md"
        content_file.write_text(
            PAGE_TEMPLATE.format(
                title=f"Page {n}",
                slug=f"page{n}",
                url="{{ site_map.find('page0', collection='coll1').url_for }}",
            )
        )

    class testSite(Site):
        output_path = tmp_output_path
        _template_path = tmp_template_path

    site = testSite()

    for collection in ["coll1", "coll2"]:

        class MyColl(Collection):
            routes = [f"{collection}-route"]
            pages = coll_pages[collection]

        collection_class = MyColl
        collection_class.__name__ = collection
        site.collection(collection_class)

    for n in range(4):

        @site.page
        class MyPage(Page):
            content_path = tmp_input_path / f"page{n}.md"
            template = template_file_name
            skip_site_map = n == 3

    yield site


def test_site_map_to_html(site):
    sm = SiteMap("", site.route_list)
    assert sm.html == (
        "<ul>\n"
        '\t<li><a href="/coll1-route">coll1</a></li>\n'
        "\t<ul>\n"
        '\t\t<li><a href="/coll1-route/page0.html">coll1 -- Page 0</a></li>\n'
        '\t\t<li><a href="/coll1-route/page1.html">coll1 -- Page 1</a></li>\n'
        '\t\t<li><a href="/coll1-route/page2.html">coll1 -- Page 2</a></li>\n'
        '\t\t<li><a href="/coll1-route/page3.html">coll1 -- Page 3</a></li>\n'
        "\t</ul>\n"
        '\t<li><a href="/coll2-route">coll2</a></li>\n'
        "\t<ul>\n"
        '\t\t<li><a href="/coll2-route/page0.html">coll2 -- Page 0</a></li>\n'
        '\t\t<li><a href="/coll2-route/page1.html">coll2 -- Page 1</a></li>\n'
        '\t\t<li><a href="/coll2-route/page2.html">coll2 -- Page 2</a></li>\n'
        '\t\t<li><a href="/coll2-route/page3.html">coll2 -- Page 3</a></li>\n'
        "\t</ul>\n"
        '\t<li><a href="/page0.html">Page 0</a></li>\n'
        '\t<li><a href="/page1.html">Page 1</a></li>\n'
        '\t<li><a href="/page2.html">Page 2</a></li>\n'
        "</ul>\n"
    )


@pytest.mark.parametrize(
    "value, params, expected",
    [
        ("page1", {}, "/page1.html"),
        ("page1", {"attr": "slug"}, "/page1.html"),
        ("page3", {"attr": "slug"}, None),
        ("page3", {"attr": "slug", "full_search": True}, "/coll1-route/page3.html"),
        ("page3", {"attr": "slug", "collection": "coll2"}, "/coll2-route/page3.html"),
        ("page1.html", {"attr": "path_name"}, "/page1.html"),
        ("page3.html", {"attr": "path_name"}, None),
        ("page3.html", {"attr": "path_name", "full_search": True}, "/coll1-route/page3.html"),
        ("page3.html", {"attr": "path_name", "collection": "coll2"}, "/coll2-route/page3.html"),
        ("page1", {"attr": "slug"}, "/page1.html"),
        ("page3", {"attr": "slug"}, None),
        ("page3", {"attr": "slug", "full_search": True}, "/coll1-route/page3.html"),
        ("page3", {"attr": "slug", "collection": "coll2"}, "/coll2-route/page3.html"),
        ("Page 1", {"attr": "title"}, "/page1.html"),
        ("coll1 -- Page 3", {"attr": "title"}, None),
        ("coll1 -- Page 3", {"attr": "title", "full_search": True}, "/coll1-route/page3.html"),
        ("coll1 -- Page 3", {"attr": "title", "collection": "coll2"}, None),
        ("coll2 -- Page 3", {"attr": "title", "collection": "coll2"}, "/coll2-route/page3.html"),
    ],
)
def test_site_map_search(site, value, params, expected):
    sm = SiteMap("", site.route_list)
    if expected is not None:
        found = sm.find(value, **params)
        assert isinstance(found, SiteMapEntry)
        assert found.url_for == expected
    else:
        assert sm.find(value, **params) is None


def test_find_in_template(site):
    site.render()
    assert (site.output_path / "page1.html").read_text() == "Page 1\n/coll1-route/page0.html"


def test_site_map_to_xml(site):
    site.render_xml_site_map = True
    site.render()
    assert (
        (site.output_path / "site_map.xml").read_text()
        == """<?xml version="1.0" encoding="UTF-8"?>
<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
<url>
	<loc>http://localhost:8000/coll1-route</loc>
</url>
<url>
	<loc>http://localhost:8000/coll1-route/page0.html</loc>
</url>
<url>
	<loc>http://localhost:8000/coll1-route/page1.html</loc>
</url>
<url>
	<loc>http://localhost:8000/coll1-route/page2.html</loc>
</url>
<url>
	<loc>http://localhost:8000/coll1-route/page3.html</loc>
</url>
<url>
	<loc>http://localhost:8000/coll2-route</loc>
</url>
<url>
	<loc>http://localhost:8000/coll2-route/page0.html</loc>
</url>
<url>
	<loc>http://localhost:8000/coll2-route/page1.html</loc>
</url>
<url>
	<loc>http://localhost:8000/coll2-route/page2.html</loc>
</url>
<url>
	<loc>http://localhost:8000/coll2-route/page3.html</loc>
</url>
<url>
	<loc>http://localhost:8000/page0.html</loc>
</url>
<url>
	<loc>http://localhost:8000/page1.html</loc>
</url>
<url>
	<loc>http://localhost:8000/page2.html</loc>
</url>
</urlset>"""
    )



from collections.abc import Generator, Iterable
from pathlib import Path
from urllib.parse import urljoin

import slugify

from render_engine import Collection, Page
from render_engine._base_object import BaseObject
from render_engine.data_object import DataObject


class SiteMapEntry:
    """Entry in the site map"""

    def __init__(self, entry: BaseObject, route: str | Path, from_collection=False):
        """Initialize the entry"""
        self.slug = entry._slug
        self.title = entry._title
        self.path_name = entry.path_name
        route = str(route)
        match entry:
            case Page() | DataObject():
                # For a base page the _route created if we use the route is invalid - just use the path_name
                self._route = f"/{route.lstrip('/')}/{self.path_name}" if from_collection else f"/{self.path_name}"
                self.entries = list()
            case Collection():
                self._route = f"/{str(entry.routes[0]).lstrip('/')}"
                self.entries = [
                    SiteMapEntry(collection_entry, self._route, from_collection=True) for collection_entry in entry
                ]
            case _:
                pass

    @property
    def url_for(self) -> str:
        """The URL for the given entry"""
        return str(self._route)

    def __str__(self) -> str:
        """String representation of the entry as its URL"""
        return self.url_for


class SiteMap:
    """Site map"""

    def __init__(self, site_url: str = "", route_list: dict | None = None) -> None:
        """
        Create the site map based on the route_list

        :param site_url: Used for rendering the HTML to have absolute URLs.
        :param route_list: The route list to parse
        """
        self._route_map = dict()
        self._collections = dict()
        self._site_url = site_url
        if not route_list:
            return
        self.update(route_list)

    @property
    def site_url(self) -> str:
        return self._site_url

    @site_url.setter
    def site_url(self, value: str) -> None:
        self._site_url = value

    def update(self, route_list: dict) -> None:
        """
        Update the site map with a new route list.

        :param route_list: The route list to parse
        """
        route: str
        entry: BaseObject
        self._route_map = dict()
        for route, entry in route_list.items():
            if entry.skip_site_map:
                continue
            sm_entry = SiteMapEntry(entry, route)
            self._route_map[sm_entry.slug] = sm_entry
            if sm_entry.entries:
                self._collections[sm_entry.slug] = sm_entry

    def __iter__(self) -> Generator[SiteMapEntry]:
        """Iterator for the site map object"""
        for entry in self._route_map.values():
            yield entry
            yield from entry.entries

    def find(
        self,
        value: str,
        attr: str = "slug",
        collection: str | None = None,
        full_search: bool = False,
    ) -> SiteMapEntry | None:
        """
        Find an entry in the site map

        Only one of the parameters slug, path_name, or title may be set.
        If collection is set that collection will be searched. If it is not set then all collections
        will be searched if full_search is True.
        If there would be a match in multiple collections - or for just a page, the first match will be returned.

        :param value: The value of attribute to search for
        :param attr: The name of attribute to search for, defaults to slug
        :param collection: The name of the collection to search
        :param full_search: Search recursively in collections
        :return: The first found match or None if not found
        """

        def search(s_attr: str, s_value: str, entries: Iterable) -> SiteMapEntry | None:
            """
            Search the list of entries for a match

            :param s_attr: The attribute to search by
            :param s_value: The value to search for
            :param entries: List of entries to search
            :return: First found match or None if not found
            """
            for s_entry in entries:
                if getattr(s_entry, s_attr, None) == s_value:
                    return s_entry
            return None

        if not value:
            return None
        if collection:
            # Collection was specified so check there
            return (
                search(attr, value, _collection.entries)
                if (_collection := self._collections[slugify.slugify(collection)])
                else None
            )
        if attr == "slug":
            # We can handle slug a bit differently since it's the key to the route map dictionary.
            if entry := self._route_map.get(value):
                return entry
        # Check the base route map
        elif entry := search(attr, value, self._route_map.values()):
            return entry
        if full_search:
            # Check each collection
            for collection in self._collections.values():
                if entry := search(attr, value, collection.entries):
                    return entry
        return None

    @property
    def html(self) -> str:
        """Build the site map as HTML"""
        html_string = "<ul>\n"
        # We can't iterate over `self` because that will flatten out the site map and we do not want that for the HTML
        # version.
        for entry in self._route_map.values():
            html_string += f'\t<li><a href="{urljoin(self.site_url, entry.url_for)}">{entry.title}</a></li>\n'
            if entry.entries:
                html_string += "\t<ul>\n"
                html_string += "".join(
                    f'\t\t<li><a href="{urljoin(self.site_url, sub_entry.url_for)}">{sub_entry.title}</a></li>\n'
                    for sub_entry in entry.entries
                )
                html_string += "\t</ul>\n"
        html_string += "</ul>\n"
        return html_string



import logging
import pathlib

import jinja2
import pytest

from render_engine import Page


@pytest.fixture
def page_from_file(tmp_path: pathlib.Path):
    d = tmp_path / "test_page.md"
    content = """---
title: Test Page
custom: "test"
---

# Test Page
This is a test page
"""

    d.write_text(content)
    return Page(content_path=d)


def test_page_attrs_from_file(page_from_file: Page):
    """
    Tests that expected page attrsibutes are set from the file.
    Currently this is handled by the BasePageParser and the logic in the Page.
    """
    assert page_from_file._title == "Test Page"


def test_page_custom_attrs_from_file(page_from_file: Page):
    """Tests that unique page attrsibutes are set from the file"""
    assert page_from_file.custom == "test"


def test_page_from_template(tmp_path: pathlib.Path):
    """Tests that page attributes are set from a template"""

    class CustomPage(Page):
        template = "test.html"
        title = "Test Page"
        template = "test.html"

    environment = jinja2.Environment(loader=jinja2.DictLoader({"test.html": "{{ title }}"}))

    page = CustomPage()
    assert page._render_content(engine=environment) == "Test Page"


def test_page_content_renders_jinja():
    """Tests that page content is rendered with jinja"""

    class CustomPage(Page):
        content = "Test Page"

    page = CustomPage()
    assert page.content == "Test Page"
    assert page._content == "Test Page"


def test_rendered_page_from_template_has_attributes():
    """Tests that selected page attributes are available in a template"""
    template = "{{title}}-{{slug}}-{{url}}"

    environment = jinja2.Environment(loader=jinja2.DictLoader({"test.html": template}))

    class CustomPage(Page):
        template = environment.get_template("test.html")

    assert CustomPage()._render_from_template(template=CustomPage.template) == "CustomPage-custompage-/custompage.html"


def test_rendered_page_from_template_has_content():
    """Tests that selected page attributes are available in a template"""
    template = "{{content}}"

    environment = jinja2.Environment(loader=jinja2.DictLoader({"test.html": template}))

    class CustomPage(Page):
        template = environment.get_template("test.html")
        content = "This is a custom page"

    assert CustomPage()._render_from_template(template=CustomPage.template) == "This is a custom page"


def test_rendered_page_from_template_has_data():
    """Tests that selected page attributes are available in a template"""
    template = "{% for d in data %}{{d}}{% endfor %}"

    environment = jinja2.Environment(loader=jinja2.DictLoader({"test.html": template}))

    class CustomPage(Page):
        template = environment.get_template("test.html")
        data = [1, 2, 3, 4]

    assert CustomPage()._render_from_template(template=CustomPage.template) == "1234"


def test_page_fails_to_render_content_as_template(caplog):
    """Tests handling content that fails to render as a template"""
    caplog.set_level(logging.INFO)

    template = "{% for d in data %}{{d}}{% endfor %}\n{{content}}"

    environment = jinja2.Environment(loader=jinja2.DictLoader({"test.html": template}))

    class CustomPage(Page):
        template = environment.get_template("test.html")
        data = [1, 2, 3, 4]
        content = "{{ site_map.find('test') }}"

    assert CustomPage()._render_from_template(template=CustomPage.template) == "1234\n{{ site_map.find('test') }}"
    assert "Failed to pre-render" in caplog.messages[0]


def test_braces_ignored_without_sitemap():
    """Tests that braces in content are ignored without `site_map`"""
    template = "{% for d in data %}{{d}}{% endfor %}\n{{content}}"

    environment = jinja2.Environment(loader=jinja2.DictLoader({"test.html": template}))

    class CustomPage(Page):
        template = environment.get_template("test.html")
        data = [1, 2, 3, 4]
        content = "{{ example }}"

    assert CustomPage()._render_from_template(template=CustomPage.template) == "1234\n{{ example }}"


def test_exception_in_parsing_content_as_template(monkeypatch, caplog):
    """Test failing to parse content as template"""

    def mock_template():
        raise jinja2.exceptions.UndefinedError("Mocked error")

    monkeypatch.setattr("render_engine.page.Template", mock_template)
    caplog.set_level(logging.INFO)

    template = "{% for d in data %}{{d}}{% endfor %}\n{{content}}"

    environment = jinja2.Environment(loader=jinja2.DictLoader({"test.html": template}))

    class CustomPage(Page):
        template = environment.get_template("test.html")
        data = [1, 2, 3, 4]
        content = "{{ site_map.find('test') }}"

    assert CustomPage()._render_from_template(template=CustomPage.template) == "1234\n{{ site_map.find('test') }}"
    assert "Failed to parse" in caplog.messages[0]


def test_no_prerender():
    """Test pre-rendering does not occur when no_prerender is True"""
    template = "{% for d in data %}{{d}}{% endfor %}\n{{content}}"

    environment = jinja2.Environment(loader=jinja2.DictLoader({"test.html": template}))

    class CustomPage(Page):
        template = environment.get_template("test.html")
        data = [1, 2, 3, 4]
        content = "{{ site_map.find('test') }}"
        no_prerender = True

    assert CustomPage()._render_from_template(template=CustomPage.template) == "1234\n{{ site_map.find('test') }}"



import logging
from collections import defaultdict
from collections.abc import Iterable
from typing import Any

import pluggy

_PROJECT_NAME = "render_engine"
hook_impl = pluggy.HookimplMarker(project_name=_PROJECT_NAME)
hook_spec = pluggy.HookspecMarker(project_name=_PROJECT_NAME)


class SiteSpecs:
    """
    Plugin hook specifications for the Site class.

    This class defines the hook specifications for various stages of site building and rendering.
    Each hook method represents a specific stage and can be implemented by plugins to customize the behavior.

    Attributes:
        default_settings (dict[str, Any]): The default settings for the site.

    """

    default_settings: dict[str, Any]

    @hook_spec
    def add_default_settings(
        self,
        site,
    ) -> None:
        """Add default settings to the site"""

    @hook_spec
    def pre_build_site(
        self,
        site,
        settings: dict[str, Any],
    ) -> None:
        """Steps Prior to Building the site"""

    @hook_spec
    def post_build_site(
        self,
        site,
        settings: dict[str, Any],
    ) -> None:
        """Build After Building the site"""

    @hook_spec
    def render_content(
        self,
        page,
        settings: dict[str, Any],
        site,
    ) -> None:
        """
        Augments the content of the page before it is rendered as output.
        """

    @hook_spec
    def post_render_content(
        self,
        page,
        settings: dict[str, Any],
        site,
    ) -> None:
        """
        Augments the content of the page before it is rendered as output.
        """

    @hook_spec
    def pre_build_collection(
        self,
        collection,
        site,
        settings: dict[str, Any],
    ) -> None:
        """Steps Prior to Building the collection"""

    @hook_spec
    def post_build_collection(
        self,
        collection,
        site,
        settings: dict[str, Any],
    ) -> None:
        """Build After Building the collection"""


class PluginManager:
    """
    Manages the plugins for the site.

    Attributes:
        plugin_settings (dict): A dictionary that stores the settings for each plugin.
    """

    plugin_settings: dict = defaultdict(dict)

    def __init__(self):
        self._pm = pluggy.PluginManager(project_name=_PROJECT_NAME)
        self._pm.add_hookspecs(SiteSpecs)

    def register_plugin(self, plugin) -> None:
        """Register a plugin with the plugin manager"""
        if self._pm.has_plugin(plugin.__name__):
            logging.info(f"Plugin {plugin} already registered")
            return
        self._pm.register(plugin)

    def unregister_plugin(self, plugin):
        """Unregister a plugin with the plugin manager"""
        if self._pm.has_plugin(plugin.__name__):
            self._pm.unregister(plugin)

    @property
    def plugins(self) -> set[Any]:
        """
        Get the set of registered plugins.

        Returns:
            set: A set containing the registered plugins.
        """
        return self._pm.get_plugins()

    @property
    def hook(self):
        """
        Give access to the hook to run plugins.

        Returns:
            HookRelay: The hook to actually trigger the plugins.
        """
        return self._pm.hook


def handle_plugin_registration(
    plugin_manager: PluginManager,
    plugins: Iterable[tuple[type, dict]],
    current_settings: dict,
):
    """
    Register plugins to a plugin_manager

    :param plugin_manager: The plugin manager to register the plugins
    :param plugins: List of tuples containing the plugin and settings:
        [(Plugin1, {}), (Plugin2, {'setting': 'updated'})]
    :param current_settings: Dictionary of current settings
    """
    for plugin_definition in plugins:
        if isinstance(plugin_definition, tuple):
            if len(plugin_definition) == 2:
                plugin, settings = plugin_definition
            else:
                plugin, settings = plugin_definition[0], dict()
        else:
            plugin, settings = plugin_definition, dict()

        plugin_name = plugin.__name__
        logging.debug(f"Registering Plugin {plugin_name}")
        plugin_manager.register_plugin(plugin)
        default_settings = getattr(plugin, "default_settings", dict())
        logging.debug(
            f"Registering settings: default: {default_settings} "
            f"current: {current_settings.get(plugin_name, dict())} "
            f"overrides: {settings}"
        )
        plugin_manager.plugin_settings[plugin_name] = {
            **default_settings,
            **current_settings.get(plugin_name, dict()),
            **settings,
        }



import pytest
from jinja2 import Environment, select_autoescape

from render_engine.collection import Collection
from render_engine.engine import (
    render_engine_templates_loader,
    to_absolute,
    to_pub_date,
)
from render_engine.page import Page
from render_engine.site import Site


@pytest.fixture
def engine() -> Environment:
    env = Environment(
        loader=render_engine_templates_loader,
        autoescape=select_autoescape(["xml"]),
        lstrip_blocks=True,
        trim_blocks=True,
    )

    env.filters["to_pub_date"] = to_pub_date
    env.filters["to_absolute"] = to_absolute
    return env


@pytest.fixture()
def site(tmp_path):
    tmp_dir = tmp_path / "content"
    tmp_dir.mkdir()
    file = tmp_dir / "#"
    file.write_text("test")

    site = Site()

    @site.collection
    class TestCollection(Collection):
        pages = [Page(content_path=file)]
        has_archive = True

    return site



import pytest

from render_engine.archive import Archive
from render_engine.collection import Collection
from render_engine.page import Page


def test_archive_slug_named_after_title():
    """Archives usually get their name from their collection. This tests that the slug is named after the title"""

    archive = Archive(
        title="test archive",
        pages=[Page()],
        template="",
        routes=["./"],
        template_vars={},
    )

    assert archive._slug == "test-archive"


@pytest.mark.parametrize(
    "title, expected_slug",
    [
        ("archive", "archive1"),
        ("collection", "collection1"),
        ("Test Collection", "test-collection1"),
    ],
)
def test_archive_slug_name_with_pages(title, expected_slug):
    """tests that if num_archive_pages is greater than 1, the slug is appended with the archive_index"""

    archive = Archive(
        title=title,
        pages=[Page()],
        template="",
        routes=["./"],
        archive_index=1,
        template_vars={},
    )

    assert archive._slug == expected_slug


@pytest.mark.parametrize(
    "_items_per_page, num_of_pages",
    [(1, 5), (2, 3), (3, 2), (4, 2), (5, 1)],
)
def test_archive_num_of_pages(_items_per_page: int, num_of_pages: int):
    """Tests that the number of pages in an archive is equal to the number of pages in the archive"""

    class TestCollection(Collection):
        pages = [Page(), Page(), Page(), Page(), Page()]
        has_archive = True
        items_per_page = _items_per_page

    collection = TestCollection()
    assert list(collection.archives)[0].template_vars["num_of_pages"] == num_of_pages

    if _items_per_page < 1:
        assert list(collection.archives)[1].template_vars["num_of_pages"] == num_of_pages
        assert list(collection.archives)[2].template_vars["num_of_pages"] == num_of_pages


def test_archive_template_is_archive_html():
    """Tests that the template is set to archive.html"""
    archive = Archive(
        title="test archive",
        pages=[Page()],
        routes=["./"],
        template_vars={},
    )
    assert archive.template == "archive.html"



import importlib
import json
import typing

import pytest

from render_engine.collection import Collection
from render_engine.page import Page
from render_engine.plugins import PluginManager, hook_impl
from render_engine.site import Site


class FakeLegacyPlugin:
    """
    Tests Plugins

    This ensures that developers can install as much of the needs they have.
    This is to check compatibility with legacy plugins.
    """

    @hook_impl
    def pre_build_collection(site):
        print("Pre Build Collection Called!")

    @hook_impl
    def post_build_collection(collection):
        print("Post Build Collection Called!")


class FakePlugin:
    """Clean the output folder before rendering"""

    default_settings = {
        "test": "default",
    }

    @hook_impl
    @staticmethod
    def pre_build_site(
        site: type[Site],
        settings: dict[str, typing.Any] | None,
    ):
        """Test Pre Build Site"""
        pass

    @hook_impl
    @staticmethod
    def render_content(
        page: Page,
        settings: dict[str, typing.Any],
    ):
        pass

    @hook_impl
    @staticmethod
    def post_render_content(
        page: Page,
        settings: dict[str, typing.Any],
    ):
        pass


@pytest.fixture(scope="module")
def site(tmp_path_factory):
    tmp_output_path = tmp_path_factory.getbasetemp() / "plugin_test_output"

    class testSite(Site):
        output_path = tmp_output_path

    site = testSite()
    site.register_plugins(FakePlugin)

    @site.collection
    class FakeCollection(Collection):
        pages = [Page(content="test")]
        has_archive = True

    @site.collection
    class IgnoredPluginCollection(Collection):
        ignore_plugins = [
            FakePlugin,
        ]
        pages = []

    @site.page
    class testPage(Page):
        ignore_plugins = [
            FakePlugin,
        ]
        content = "test"

    yield site


def test_plugin_is_registered(site: Site):
    """Check that the plugin is registered"""
    assert [site.plugin_manager._pm.get_name(x) for x in site.plugin_manager.plugins] == ["FakePlugin"]


def test_pages_in_collection_inherit_pugins(site: Site):
    """Check that collection plugins are inherited by pages in the collection"""

    assert site.route_list["fakecollection"].plugin_manager._pm.list_name_plugin()[0][0] == "FakePlugin"


def test_page_ignores_plugin(site: Site):
    """Check that the plugin is not registered in the page if it is ignored"""

    assert site.route_list["testpage"].plugin_manager._pm.list_name_plugin() == []


def test_collection_ignores_plugin(site):
    """Check that the plugin is not registered in the collection if it is ignored"""
    assert site.route_list["ignoredplugincollection"].plugin_manager._pm.list_name_plugin() == []


def test_collection_archive_inherits_plugins(site: Site, mocker):
    """Check that the archive inherits the pludef test_plugin_render_content_runs_from_archive(site, mocker):
    gins from the collection"""

    assert site.route_list["fakecollection"].plugin_manager._pm.list_name_plugin()[0][0] == "FakePlugin"


def test_collection_archive_runs_render_content_calls(site, mocker):
    mocker_render_content = mocker.patch.object(
        list(site.route_list["fakecollection"].archives)[0].plugin_manager._pm.hook,
        "render_content",
    )
    mocker_post_render_content = mocker.patch.object(
        list(site.route_list["fakecollection"].archives)[0].plugin_manager._pm.hook,
        "post_render_content",
    )

    site.render()
    assert mocker_render_content.called
    assert mocker_post_render_content.called


def test_collection_runs_render_pre_and_post_build_plugins(site, mocker):
    mocker_render_content = mocker.patch.object(
        site.route_list["fakecollection"].plugin_manager._pm.hook,
        "pre_build_collection",
    )
    mocker_post_render_content = mocker.patch.object(
        site.route_list["fakecollection"].plugin_manager._pm.hook,
        "post_build_collection",
    )

    site.render()
    assert mocker_render_content.called
    assert mocker_post_render_content.called


def test_page_plugins_registered():
    app = Site()

    @app.page
    class TestPage(Page):
        plugins = [FakePlugin]

    assert app.route_list["testpage"].plugin_manager._pm.list_name_plugin()[0][0] == "FakePlugin"


def test_deperecated_warning():
    import warnings

    with warnings.catch_warnings(record=True) as w:
        importlib.import_module("render_engine.hookspecs")
        assert any(isinstance(i.message, DeprecationWarning) for i in w)


def test_plugin_multiple_plugins():
    """Asserts that if a plugin is registered multiple times, it is only registered once"""

    plugin_mgr = PluginManager()
    plugin_mgr.register_plugin(FakePlugin)
    plugin_mgr.register_plugin(FakePlugin)


def test_plugin_override_settings_from_site_register_plugins():
    """Asserts that the settings passed into `register_plugins` override the default settings"""

    site = Site()
    site.register_plugins(FakePlugin, FakePlugin={"test": "override"})
    assert site.plugin_manager.plugin_settings.get("FakePlugin")["test"] == "override"


def test_collection_default_empty_plugin_setting(site: Site):
    """Asserts that the collection default plugin settings are set to an empty dict"""

    assert site.route_list["fakecollection"].plugin_settings == {"plugins": {}}


def test_collection_override_default_plugin_setting(site: Site):
    """Asserts that the collection default plugin settings are overridden"""

    site.route_list["fakecollection"].plugin_settings = {"FakePlugin": {"test2": "override2"}}
    assert site.route_list["fakecollection"].plugin_settings.get("FakePlugin") == {"test2": "override2"}


def test_collection__run_collection_plugins_can_handle_optional_plugins(tmp_path, capsys):
    """
    Tests that you can run
    the specific calls for modules

    You can omit args but you can't include ones that are not provided in the spec.

    `pluggy._manager.PluginValidationError: Plugin 'FakeLegacyPlugin' for hook 'pre_build_collection'
     hookimpl definition: pre_build_collection(a)
     Argument(s) {'a'} are declared in the hookimpl but can not be found in the hookspec
    `
    """

    _output_path = tmp_path / "plugins_optional_plugins"

    class TestSite(Site):
        output_path = _output_path

    site = TestSite()

    class Page1(Page):
        content = "this is a page"

    plugin_mgr = PluginManager()
    plugin_mgr.register_plugin(FakeLegacyPlugin)

    class LegacyPluginCollection(Collection):
        pages = [Page1]
        plugin_manager = plugin_mgr

    collection = LegacyPluginCollection()

    collection._run_collection_plugins(site, "pre_build_collection")
    collection._run_collection_plugins(site, "post_build_collection")

    # Capture the output
    captured = capsys.readouterr()

    # Verify the expected output was printed
    assert "Pre Build Collection Called!" in captured.out
    assert "Post Build Collection Called!" in captured.out


class TestPlugin:
    default_settings = {"settings": "default"}

    @staticmethod
    @hook_impl
    def pre_build_site(site: Site, settings: dict):
        class PreBuild(Page):
            content = json.dumps(settings.get("TestPlugin"))
            name = "prebuild"

        print("PreBuild")
        prebuild = PreBuild()
        prebuild.site = site
        prebuild.render(site.theme_manager)

    @staticmethod
    @hook_impl
    def post_build_site(site: Site, settings: dict):
        class PostBuild(Page):
            content = json.dumps(settings.get("TestPlugin"))
            name = "postbuild"

        print("PostBuild")
        postbuild = PostBuild()
        postbuild.site = site
        postbuild.render(site.theme_manager)

    @staticmethod
    @hook_impl
    def render_content(site: Site, settings: dict):
        class RenderContent(Page):
            content = json.dumps(settings.get("TestPlugin"))
            name = "renddercontent"

        print("RenderContent")
        renddercontent = RenderContent()
        renddercontent.site = site
        renddercontent.render(site.theme_manager)

    @staticmethod
    @hook_impl
    def post_render_content(site: Site, settings: dict):
        class PostRenderContent(Page):
            content = json.dumps(settings.get("TestPlugin"))
            name = "postrendercontent"

        print("PostRenderContent")
        postrendercontent = PostRenderContent()
        postrendercontent.site = site
        postrendercontent.render(site.theme_manager)

    @staticmethod
    @hook_impl
    def pre_build_collection(site: Site, settings: dict):
        class PreBuildCollection(Page):
            content = json.dumps(settings.get("TestPlugin"))
            name = "prebuildcollection"
            site = site

        print("PreBuildCollection")
        prebuildcollection = PreBuildCollection()
        prebuildcollection.site = site
        prebuildcollection.render(site.theme_manager)

    @staticmethod
    @hook_impl
    def post_build_collection(site: Site, settings: dict):
        class PostBuildCollection(Page):
            content = json.dumps(settings.get("TestPlugin"))
            name = "postbuildcollection"

        print("PostBuildCollection")
        postbuildcollection = PostBuildCollection()
        postbuildcollection.site = site
        postbuildcollection.render(site.theme_manager)


@pytest.fixture(scope="function")
def plugin_test_site(tmp_path):
    """Fixture for site to test plugins"""
    _output_path = tmp_path / "plugin_test_site"

    class TestSite(Site):
        output_path = _output_path

    yield TestSite()


@pytest.mark.parametrize(
    "settings, expected",
    [
        ({}, TestPlugin.default_settings),
        ({"settings": "override"}, {"settings": "override"}),
        ({"new_setting": "new"}, {"new_setting": "new", **TestPlugin.default_settings}),
    ],
)
def test_plugin_settings_are_passed_properly(plugin_test_site, settings, expected):
    """Test that plugins are properly passed"""

    class Page1(Page):
        content = "this is a page"
        site = plugin_test_site

    @plugin_test_site.collection
    class LegacyPluginCollection(Collection):
        pages = [Page1()]
        plugins = [(TestPlugin, settings)] if settings else [TestPlugin]

    plugin_test_site.register_plugins(TestPlugin, TestPlugin=settings)

    plugin_test_site.render()
    for hook in [
        "prebuild",
        "postbuild",
        "prebuildcollection",
        "postbuildcollection",
        "rendercontent",
        "postrendercontent",
    ]:
        path = plugin_test_site.output_path / f"{hook}.html"
        with open(path) as f:
            assert json.load(f) == expected, f"Failed to match {settings=} to {expected=}"



import json
from pathlib import Path

import pluggy
import pytest
import toml
from jinja2 import DictLoader, FileSystemLoader

from render_engine import DataObject
from render_engine.collection import Collection
from render_engine.page import Page
from render_engine.plugins import SiteSpecs
from render_engine.site import Site
from render_engine.themes import Theme

pm = pluggy.PluginManager("fake_test")


@pytest.fixture
def site(tmp_path: Path):
    _site = Site()
    _site.output_path = tmp_path / "output"
    _site.static_paths.add(tmp_path / "static")
    return _site


def test_site_defaults(site):
    """
    Tests that a site can be created with default variables.
    Those values are:
        - SITE_TITLE: "Untitled Site"
        - SITE_URL: "http://example.com"
    """

    assert site.site_vars["SITE_TITLE"] == "Untitled Site"
    assert site.site_vars["SITE_URL"] == "http://localhost:8000/"


def test_site_site_vars_override_defaults_via_class():
    """
    Tests that a site can be created with default variables.
    """

    site_vars = {
        "SITE_TITLE": "My Site",
        "SITE_URL": "https://my-site.com",
    }

    site.site_vars = site_vars

    assert site.site_vars["SITE_TITLE"] == "My Site"
    assert site.site_vars["SITE_URL"] == "https://my-site.com"


def test_site_page_in_route_list(site, tmp_path: Path):
    tmp_dir = tmp_path / "content"
    tmp_dir.mkdir()
    file = tmp_dir / "test.md"
    file.write_text("test")

    # assert that the route list is empty
    assert len(site.route_list) == 0

    class CustomPage(Page):
        test_value = "test"
        content_path = file.absolute()

    site.page(CustomPage)

    assert site.route_list["custompage"].test_value == "test"


def test_site_collection_in_route_list(site):
    """Tests that when a collection is added to the route_list it is only the colleciton"""
    # assert that the route list is empty
    assert len(site.route_list) == 0

    class CustomPage1(Page):
        pass

    class CustomPage2(Page):
        pass

    class collection(Collection):
        pages = [CustomPage1(), CustomPage2()]

    collection = site.collection(collection)

    assert site.route_list["collection"] == collection
    assert len(site.route_list) == 1
    assert "custompage1" in [getattr(page, page._reference) for page in site.route_list["collection"]]


def test_site_page_with_multiple_routes_has_one_entry_in_routes_list(site):
    """Tests a page with multiple routes only has one entry in the routes list"""

    class CustomPage(Page):
        test_value = "test"
        routes = ["customroute", "customroute2"]

    site.page(CustomPage)

    assert len(site.route_list) == 1


def test_url_for_Page_in_site(site, tmp_path: Path):
    """Tests that url_for a page is added to a template"""
    test_template = Path(tmp_path / "template.html")
    test_template.write_text("The URL is '{{ 'custompage'|url_for }}'")
    site.theme_manager.engine.loader.loaders.insert(0, FileSystemLoader(tmp_path))

    @site.page
    class CustomPage(Page):
        template = test_template.name

    site.render()
    custom_page = site.output_path / "custompage.html"
    assert custom_page.exists()
    assert custom_page.read_text() == "The URL is '/custompage.html'"


def test_collection_archive_in_route_list(site, tmp_path: Path):
    """Given a collection with an archive, the archive should be in the route list and accessible with url_for"""
    test_collection_archive_template = Path(tmp_path / "archive_template.html")
    test_collection_archive_template.write_text("This is the collection archive")

    test_collection_template = Path(tmp_path / "collection_archive_item_template.html")
    test_collection_template.write_text("The collection archive route is at '{{ 'customcollection' |url_for }}'")

    site.theme_manager.engine.loader.loaders.insert(0, FileSystemLoader(tmp_path))

    class CustomCollectionPage(Page):
        template = test_collection_template.name

    @site.collection
    class CustomCollection(Collection):
        archive_template = test_collection_archive_template.name
        has_archive = True
        pages = [CustomCollectionPage()]

    site.render()
    assert site.output_path.joinpath("customcollection.html").exists()
    assert site.output_path.joinpath("customcollectionpage.html").exists()
    assert site.output_path.joinpath("customcollection.html").read_text() == "This is the collection archive"
    assert (
        site.output_path.joinpath("customcollectionpage.html").read_text()
        == "The collection archive route is at '/customcollection.html'"
    )


@pytest.fixture(scope="module")
def site_with_collection(tmp_path_factory: pytest.TempPathFactory):
    collection_archive_path = tmp_path_factory.getbasetemp() / "collection_archive_items"
    collection_archive_output_path = collection_archive_path / "output"
    static_tmp_dir = collection_archive_path / "static"
    static_tmp_dir.mkdir(parents=True)
    Path(static_tmp_dir / Path("test.txt")).write_text("test")

    _site = Site()
    _site.output_path = collection_archive_output_path
    _site.static_paths.add(static_tmp_dir)

    class CustomCollectionPages1(Page):
        content = "test"

    class CustomCollectionPages2(Page):
        content = "test"

    @_site.collection
    class CustomPagesCollection(Collection):
        has_archive = True
        pages = [CustomCollectionPages1(), CustomCollectionPages2()]
        items_per_page = 1

    _site.render()
    return _site


def test_collection_archives_generates_by_items_per_page(site_with_collection: Site):
    """
    Archive pages should be created using the items_per_page value

    Example:
        If items_per_page is 1, and there are 2 pages then there should be 3 archive pages.
        0: all page items
        1: first page item
        2: second page item
    """

    assert len(list(site_with_collection.route_list["custompagescollection"].archives)) == 3


def test_collection_archive_pages_in_route_list(site_with_collection: Site):
    """Given a collection with an archive, the archive should be in the route list and accessible with url_for"""

    for page in site_with_collection.route_list["custompagescollection"].archives:
        assert Path(site_with_collection.output_path / page.path_name).exists()


def test_url_for_Collection_in_site(site: Site, tmp_path: Path):
    """
    Tests that url_for a page in a collection is added to a template
    """
    test_template = Path(tmp_path / "custom_template.html")
    test_template.write_text("The URL is '{{ 'customcollection.customcollectionpage' | url_for }}'")
    if site.theme_manager.engine.loader is not None:
        site.theme_manager.engine.loader.loaders.insert(0, FileSystemLoader(tmp_path))

    class CustomCollectionPage(Page):
        template = test_template.name

    @site.collection
    class CustomCollection(Collection):
        template = test_template.name
        pages = [CustomCollectionPage()]

    site.render()
    custom_page = site.output_path / "customcollectionpage.html"
    assert custom_page.exists()
    assert custom_page.read_text() == "The URL is '/customcollectionpage.html'"


def test_site_output_path(site, tmp_path: Path):
    """Tests site outputs to output_path"""

    @site.page
    class CustomPage(Page):
        content = "this is a test"

    site.render()

    assert (site.output_path / "custompage.html").exists()


def test_site_static_renders_in_static_output_path(site_with_collection: Site):
    """
    Tests that a static file is rendered in the static output path.
    """

    assert Path(site_with_collection.output_path, "static", "test.txt").exists()


def tests_site_nested_static_paths(tmp_path: Path, site: Site):
    """given a static path with nested directories, the output should be the same"""
    static_tmp_dir = tmp_path / "static"
    output_tmp_dir = tmp_path / "output"
    output_tmp_dir.mkdir()
    static_tmp_dir.mkdir()
    Path(static_tmp_dir / "nested").mkdir()
    Path(static_tmp_dir / "nested" / Path("test.txt")).write_text("test")
    Path(static_tmp_dir / Path("test.txt")).write_text("test")

    site.output_path = output_tmp_dir
    site.static_paths.add(static_tmp_dir)
    site.render()
    assert (output_tmp_dir / "static" / "test.txt").exists()
    assert (output_tmp_dir / "static" / "nested" / "test.txt").exists()


def tests_site_multiple_static_paths(tmp_path: Path, site: Site):
    """given a static path with nested directories, the output should be the same"""
    static_tmp_dir = tmp_path / "static"
    output_tmp_dir = tmp_path / "output"
    output_tmp_dir.mkdir()
    static_tmp_dir.mkdir()
    Path(static_tmp_dir / Path("test.txt")).write_text("test")

    second_static_tmp_dir = tmp_path / "static2"
    second_static_tmp_dir.mkdir()
    Path(second_static_tmp_dir / Path("test2.txt")).write_text("test")

    site.output_path = output_tmp_dir
    site.static_paths.update([static_tmp_dir, second_static_tmp_dir])
    site.render()
    assert (output_tmp_dir / "static" / "test.txt").exists()
    assert (output_tmp_dir / "static2" / "test2.txt").exists()


def test_site_theme_update_settings(site, tmp_path: Path):
    """Tests that the theme manager updates the settings"""
    site = Site()

    assert "theme" in site.site_vars
    assert "test" not in site.site_vars
    site.update_theme_settings(test="test")
    assert site.site_vars["theme"]["test"] == "test"


def test_plugin_in_theme_added_to_plugins(site):
    """Tests that a plugin added to a theme is added to the site"""

    class plugin(SiteSpecs):
        pass

    class theme(Theme):
        loader = DictLoader({"test.html": "test"})
        prefix = "test"
        plugins = [plugin]
        filters = []

    site.register_theme(
        theme,
    )
    assert plugin in site.plugin_manager.plugins


def test_collection_archive_0_is_index(site, tmp_path: Path):
    test_collection_archive = Path(tmp_path / "archive_template")
    site.output_path = test_collection_archive

    class CustomCollectionPage(Page):
        content = "Test"

    @site.collection
    class CustomCollection(Collection):
        routes = ["test"]
        has_archive = True
        pages = [CustomCollectionPage()]

    site.render()
    assert Path(test_collection_archive / "test" / "index.html").exists()


def test_custom_template_path_assignment(site, tmp_path):
    """
    asserts that you can assign a new template path to an initialized site

    Template paths are used to find templates when rendering pages.

    The attribute is saved in the template_manager's ChoiceLoader.
    """

    test_custom_template_path = Path(tmp_path / "custom_template_path")
    test_custom_template_path.mkdir()
    test_custom_template = test_custom_template_path / "test_custom_template.html"
    test_custom_template.touch()
    site = Site()
    site.template_path = str(test_custom_template_path)
    assert site.theme_manager.engine.get_template("test_custom_template.html")


@pytest.mark.parametrize(
    "seri_out, seri_in, kwargs",
    [
        (json.dumps, json.loads, None),
        (json.dumps, json.loads, {}),
        (json.dumps, json.loads, {"indent": 2}),
        (toml.dumps, toml.loads, None),
        (None, json.loads, None),
    ],
)
def test_site_with_data_object(tmp_path, seri_in, seri_out, kwargs):
    """Tests that the data object renders properly"""
    site = Site()
    output_path: Path = Path(tmp_path)
    site.output_path = output_path

    source_data: dict = {"key": "value"}
    output_filename = "data_object.json"

    @site.data_object
    class MyDataObject(DataObject):
        data_object = source_data
        path_name = output_filename
        serializer = seri_out

    site.render()
    output_file = (output_path / output_filename).read_text()
    assert seri_in(output_file) == source_data



import copy
import datetime
import logging
import os
from collections.abc import Callable, Generator, Iterable
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Any, cast

import dateutil.parser as dateparse
from more_itertools import batched
from render_engine_parser import BasePageParser
from slugify import slugify

from ._base_object import BaseObject
from .archive import Archive
from .content_managers import ContentManager, FileContentManager
from .feeds import RSSFeed
from .page import Page
from .plugins import PluginManager


class Collection(BaseObject):
    """
    Collection objects serve as a way to quickly process pages that have a
    portion of content that is similar or file driven.

    Example:

    ```python
    from render_engine import Site, Collection

    site = Site()

    @site.collection
    class BasicCollection(Collection):
        content_path = "content/pages"
    ```

    Collection pages **MUST** come from a `content_path` and all be the same
    content type.  `content_path` can be a string representing a path or URL,
    depending on the [parser][src.render_engine.parsers.base_parsers] used.

    Attributes:

        archive_template: The template to use for the [`Archive`][src.render_engine.archive.Archive] pages.
        content_path: The path to iterate over to generate pages.
        content_type: Type[Page] = Page
        Feed: Type[RSSFeed] = RSSFeed
        feed_title: str
        include_suffixes: list[str] = ["*.md", "*.html"]
        items_per_page: int | None
        Parser: BasePageParser = BasePageParser
        parser_extras: dict[str, Any]
        required_themes: list[callable]
        routes: list[str | Path] = ["./"]
        sort_by: str | list = "title"
        sort_reverse: bool = False
        title: str
        template: str | None
        archive_template str | None: The template to use for the archive pages.
        ContentManager: type[ContentManager] | None = FileContentManager
        content_manager: ContentManager
        content_manager_extras: dict[str, Any]: kwargs to pass to the ContentManager when instantiating

    Methods:

        iter_content_path(): Iterates through the collection's content path.
        get_page(content_path: str | Path | None = None): Returns the page Object for the specified Content Path.
        sorted_pages: Returns the sorted pages of the collection.
        archives: Returns the Archive objects containing the pages from the content path.
        feed: Returns the Feed object for the collection.
        slug: Returns the slugified title of the collection.

    """

    archive_template: str | Path | None = "archive.html"
    content_path: Path | str
    content_type: type[Page] = Page
    Feed: type[RSSFeed] = RSSFeed
    feed_title: str
    include_suffixes: list[str] = ["*.md", "*.html"]
    items_per_page: int | None
    Parser: type[BasePageParser] = BasePageParser
    parser_extras: dict[str, Any]
    required_themes: list[Callable]
    routes: list[str | Path] = ["./"]
    site = None
    sort_by: str | list = "_title"
    sort_reverse: bool = False
    template_vars: dict[str, Any]
    template: str | None
    plugin_manager: PluginManager | None
    ContentManager: type[ContentManager] = FileContentManager
    content_manager_extras: dict[str, Any]

    def __init__(
        self,
    ) -> None:
        if parser := getattr(self, "PageParser", None):
            logging.warning(
                DeprecationWarning(
                    f"The deprecated`PageParser` attribute is used in `{self.__class__.__name__}`. \
                        Use the `Parser` attribute instead."
                )
            )
            self.Parser = parser

        if getattr(self, "items_per_page", False):
            self.has_archive = True
        self.title = self._title
        self.template_vars = getattr(self, "template_vars", {})

        cm_extras = {
            "content_path": getattr(self, "content_path", None),
            "include_suffixes": getattr(self, "include_suffixes", None),
            "collection": self,
        }
        if hasattr(self, "content_manager_extras"):
            cm_extras.update(self.content_manager_extras)
        self.content_manager = self.ContentManager(**cm_extras)
        if hasattr(self, "pages"):
            self.content_manager.pages = cast(Iterable, self.pages)

    def get_page(
        self,
        content_path: str | Path | None = None,
    ) -> Page:
        """Returns the page Object for the specified Content Path"""
        _page = self.content_type(
            content_path=content_path,
            Parser=self.Parser,
        )

        _page.parser_extras = getattr(self, "parser_extras", {})
        _page.routes = self.routes
        _page.template = getattr(self, "template", None)
        _page.collection = self.to_dict()

        return _page

    @staticmethod
    def _date_key(page: Page) -> datetime.datetime:
        """
        Key for ensuring proper handling of dates for sorting
        There are 2 issues here:
        1. We need to convert a string to a datetime for proper sorting
        2. We need to strip the timezone so that we are consistently dealing with timezone naive objects

        :param page: The Page object to handle the date for
        :return: Timezone naive datetime object
        """
        date = getattr(page, "date")
        _date = dateparse.parse(date) if isinstance(date, str) else copy.copy(date)
        return _date.replace(tzinfo=None) if isinstance(_date, datetime.datetime) else _date

    def _sort_key(self, key: str | list[str]) -> Callable:
        """
        Dynamically generate the sorting key

        :param key: The key to use
        """
        if isinstance(key, str):
            return self._date_key if key == "date" else lambda page: getattr(page, key)
        return lambda page: [getattr(page, attr) for attr in key]

    @property
    def sorted_pages(self):
        """
        Returns pages in the collection sorted by the `self.sort_by` attribute.

        Exceptions:
            AttributeError: This is raised when the attribute is missing from one or more pages
            TypeError: This happens when the values being compared are of two different types

        """
        try:
            return sorted(
                (page for page in self),
                key=self._sort_key(self.sort_by),
                reverse=self.sort_reverse,
            )
        except AttributeError as e:
            raise AttributeError(
                f"Cannot sort pages: '{self.sort_by}' attribute is missing from one or more pages."
                f"Make sure all pages in collection '{self._title}' have the '{self.sort_by}' attribute defined."
            ) from e

    @property
    def archives(self) -> Generator[Archive, None, None]:
        """
        Returns a [Archive][src.render_engine.archive.Archive] objects containing the pages from the `content_path` .

        Archives are an iterable and the individual pages are built shortly after the collection pages are built.
        This happens when [Site.render][render_engine.Site.render] is called.
        """
        if not getattr(self, "has_archive", False):
            logging.warning(
                "`has_archive` is set to `False` for %s. While an archive will be generated. \
                The file will not be saved.",
                self._title,
            )
            yield from ()

        sorted_pages = list(self.sorted_pages)
        items_per_page = getattr(self, "items_per_page", len(sorted_pages))
        archives = [sorted_pages]

        if items_per_page != len(sorted_pages):
            paginated_archives = list(batched(sorted_pages, items_per_page))
            archives.extend(paginated_archives)
            self.template_vars["num_of_pages"] = len(paginated_archives)
        else:
            self.template_vars["num_of_pages"] = 1

        for index, pages in enumerate(archives):
            yield Archive(
                pages=pages,
                template=getattr(self, "archive_template", None),
                template_vars=self.template_vars,
                title=self._title,
                routes=self.routes,
                archive_index=index,
                plugin_manager=getattr(self, "plugin_manager", None),
                is_index=not index,
            )

    @property
    def feed(self):
        feed = self.Feed()
        feed.pages = self.sorted_pages
        feed.title = getattr(self, "feed_title", self._title)
        feed.slug = self._slug
        feed.Parser = self.Parser
        return feed

    @property
    def slug(self):
        return slugify(self.title)

    def __repr__(self):
        return f"{self}: {__name__}"

    def __str__(self):
        return f"{__name__}"

    def __iter__(self):
        yield from self.content_manager

    def __len__(self):
        return len(self.content_manager)

    @property
    def all_content(self) -> Generator:
        """All of the content that is associated with the Collection including Pages, Feed, and Archives"""
        yield from self

        if getattr(self, "has_archive", False):
            for archive in self.archives:
                yield archive
                if archive.is_index:
                    # In order to avoid collision with parallel processing we need to do a copy.
                    # A deepcopy is not necessary because we only care about not overwriting the
                    # slug on the original.
                    index = copy.copy(archive)
                    index.slug = "index"
                    yield index

        if feed := getattr(self, "feed", None):
            yield feed

    def _run_collection_plugins(self, site, hook_type: str):
        """
        Run plugins for a collection

        :param site: The site object triggering the call
        :param hook_type: The hook to run
        """
        self.plugin_manager = cast(PluginManager, self.plugin_manager)
        if not getattr(self.plugin_manager, "_pm", None) or not self.plugin_manager.plugins:
            return
        try:
            method = getattr(self.plugin_manager.hook, hook_type)
        except AttributeError:
            logging.error(f"Unknown {hook_type=}")
            return
        method(collection=self, site=site, settings=self.plugin_manager.plugin_settings)

    def _render(self, entry: BaseObject):
        """
        Renders 1 entry in the Collection

        :param entry: The entry to process
        """
        if not isinstance(entry, RSSFeed) and not isinstance(entry, Archive):
            entry.plugin_manager: PluginManager = copy.deepcopy(self.plugin_manager)

        # Circular imports. Need to be handled here.
        from .page import BasePage
        from .site import Site

        entry = cast(BasePage, entry)
        self = cast(Collection, self)
        self.site = cast(Site, self.site)
        entry.site = self.site
        entry.render(self.site.theme_manager)

    def render(self) -> None:
        """Iterate through Pages and Check for Archives and Feeds"""

        # Use a ThreadPool to process all of the entries in the collection in parallel.
        # This is limited to the number of CPUs available. The easiest way to implement
        # this parallelization for a single task is to use the imap_unordered method of
        # the ThreadPool. This is, effectively, a generator so we need to loop over it
        # for them to run. Since there is no actual return value to look at we just `pass`.
        for entry in ThreadPool(processes=os.cpu_count()).imap_unordered(self._render, self.all_content):
            pass

    def create_entry(
        self,
        filepath: Path | None = None,
        editor: str | None = None,
        content: str | None = None,
        metadata: dict | None = None,
    ) -> str:
        """
        Create a new entry for the Collection

        :param filepath: Path object for the new entry
        :param editor: Editor to open to edit the entry.
        :param content: Content for the new entry
        :param metadata: Metadata for the new entry
        """
        context = copy.deepcopy(self._metadata_attrs())
        if metadata:
            context.update(metadata)
        return self.content_manager.create_entry(
            filepath=filepath, editor=editor, metadata=context, content=content or "Hello, world!"
        )


def render_archives(archive, **kwargs) -> list[Archive]:
    return [archive.render(pages=archive.pages, **kwargs) for archive in archive]



from render_engine._base_object import BaseObject


class TestObject(BaseObject):
    pass


class TestBaseObjectProperties:
    test_object = TestObject()

    def test_base_object__title_defaults_to_class_name(self):
        """assert that the title defaults to the class name"""

        assert self.test_object._title == "TestObject"

    def test_base_object_plugin_settings_defaults_to_plugins_dict(self):
        """assert that the plugin_settings defaults to plugins dict"""

        assert self.test_object.plugin_settings == {"plugins": {}}

    def test_base_object__slug_defaults_to_title(self):
        """assert that the slug defaults to a slugified title"""
        assert self.test_object._slug == "testobject"

    def test_base_object__extension_defaults_to_html(self):
        """
        asserts that the extension defaults to .html
        This is to assure that is a page is generated from any object
        it will have a valid extension.
        """

        assert self.test_object.extension == ".html"

    def test_base_object__extension_starts_with_dot(self):
        """
        asserts that the extension starts with a dot
        This is to assure that is a page is generated from any object
        """

        no_dot_BaseObject = BaseObject()
        no_dot_BaseObject.extension = "xml"
        assert no_dot_BaseObject.extension == ".xml"

    def test_base_object_path_name(self):
        """Tests that the path name is returned correctly"""
        assert self.test_object.path_name == "testobject.html"

    def test_base_object_to_dict(self):
        """Tests that the to_dict method returns a dict of the object's attributes"""
        assert self.test_object.to_dict() == {
            "title": "TestObject",
            "slug": "testobject",
            "url": None,
            "path_name": "testobject.html",
            "plugins": {},
        }

    def test_base_object_to_dict_with_template_vars(self):
        """Tests that the `to_dict` adds template_vars extracted method returns a dict of the object's attributes"""
        self.test_object.template_vars = {"test": "test"}
        self.test_object.plugin_settings = {"plugins": {"test_plugin": "test"}}

        assert self.test_object.to_dict() == {
            "title": "TestObject",
            "slug": "testobject",
            "url": None,
            "test": "test",
            "template_vars": {"test": "test"},
            "path_name": "testobject.html",
            "plugin_settings": {"plugins": {"test_plugin": "test"}},
            "plugins": {"test_plugin": "test"},
        }

        assert self.test_object.template_vars == {"test": "test"}
        assert self.test_object.plugin_settings == {"plugins": {"test_plugin": "test"}}


def test_base_object():
    assert BaseObject._metadata_attrs()["title"] == "Untitled Entry"



# ruff: noqa: F821

import warnings
from typing import Any

import pluggy

from render_engine.collection import Collection
from render_engine.page import Page
from render_engine.site import Site

_PROJECT_NAME = "render_engine"
hook_impl = pluggy.HookimplMarker(project_name=_PROJECT_NAME)
warnings.warn(
    "`render_engine.hookspecs` will be deprecated in version 2024.3.1,  Please use `render_engine.plugins` instead.",
    DeprecationWarning,
)
hook_spec = pluggy.HookspecMarker(project_name=_PROJECT_NAME)


class SiteSpecs:
    """
    `hookspecs.SiteSpecs` will be deprecated in version 2024.3.1,  Please use `plugins.SiteSpecs` instead.

    Plugin hook specifications for the Site class.

    This class defines the hook specifications for various operations related to the Site class.
    These hooks can be implemented by plugins to customize the behavior of the site building process.
    """

    default_settings: dict[str, Any]

    @hook_spec
    def add_default_settings(
        self,
        site: Site,
    ) -> None:
        """Add default settings to the site"""

    @hook_spec
    def pre_build_site(
        self,
        site: Site,
        settings: dict[str, Any],
    ) -> None:
        """Steps Prior to Building the site"""

    @hook_spec
    def post_build_site(
        self,
        site: Site,
    ) -> None:
        """Build After Building the site"""

    @hook_spec
    def render_content(
        self,
        page: Page,
        settings: dict[str, Any],
    ) -> None:
        """
        Augments the content of the page before it is rendered as output.
        """

    @hook_spec
    def post_render_content(
        self,
        page: Page,
        settings: dict[str, Any],
        site: "Site",
    ) -> None:
        """
        Augments the content of the page before it is rendered as output.
        """

    @hook_spec
    def pre_build_collection(
        self,
        site: "Site",
        collection: "Collection",
        settings: dict[str, Any],
    ) -> None:
        """Steps Prior to Building the collection"""

    @hook_spec
    def post_build_collection(
        self,
        site: "Site",
        collection: "Collection",
        settings: dict[str, Any],
    ) -> None:
        """Build After Building the collection"""



"""
Feed Objects for Generating RSS Feeds
"""

from render_engine_parser import BasePageParser

from .page import BasePage


class RSSFeed(BasePage):
    """
    Creates an RSS feed [`Page`][render_engine.Page] Object.

    !!! Note
        This is the base object type and should only contain the params identified
        by the standards defined in the [RSS 2.0 Specification](http://www.rssboard.org/rss-specification)

    This is built using the built-in `rss2.0.xml` jinja template.

    !!! Note
        Some browsers may not support the `rss` extension.
        If you are having issues with your feed, try changing the extension to `xml`.

        ```python
        from render_engine import Site, RSSFeed

        class MyFeed(RSSFeed):
            extension = "xml"

        site = Site()

        @site.collection
        class MyCollection(Collection):
            Feed = MyFeed
        ```
    """

    template = "rss2.0.xml"
    extension: str = ".rss"
    Parser: type[BasePageParser] = BasePageParser

    def __init__(self):
        self.pages: list = list()
        self.slug: str = self._slug or self.__class__.__name__



import pytest
from jinja2 import DictLoader

from render_engine.page import Page
from render_engine.site import Site
from render_engine.themes import Theme


@pytest.fixture()
def theme_site(tmp_path):
    output_path = tmp_path / "output"
    output_path.mkdir()

    site = Site()
    site.output_path = output_path

    @site.page
    class TestPage(Page):
        template = "base.html"

    return site


def test_base_html_body_class(theme_site):
    """body class is added to base html template"""

    class BodyClassTheme(Theme):
        template_globals = {"body_class": "my-class"}
        loader = DictLoader({})
        filters = {}
        plugins = []
        prefix = "test"

    theme_site.register_themes(BodyClassTheme)
    theme_site.route_list["testpage"].site = theme_site
    theme_site.route_list["testpage"].render(theme_site.theme_manager)

    assert '<body class="my-class">' in (theme_site.output_path / "testpage.html").read_text()


def test_base_html_head_include(theme_site):
    """head can be imported into base html template"""

    class HeadIncludeTheme(Theme):
        template_globals = {"head": "head.html"}
        loader = DictLoader({"head.html": "<script>console.log('test')</script>"})
        filters = set()
        plugins = []
        prefix = "test"

    theme_site.register_themes(HeadIncludeTheme)
    theme_site.load_themes()
    theme_site.route_list["testpage"].site = theme_site
    theme_site.route_list["testpage"].render(theme_site.theme_manager)

    assert "<script>console.log('test')</script>" in (theme_site.output_path / "testpage.html").read_text()


def test_base_html_head_reload_theme_count(theme_site):
    """Tests that the theme is only registered once"""

    class headIncludeTheme(Theme):
        template_globals = {"head": "head.html"}
        loader = DictLoader({"head.html": "<script>console.log('test')</script>"})
        filters = set()
        plugins = []
        prefix = "test"

    theme_site.register_themes(headIncludeTheme)
    theme_site.register_themes(headIncludeTheme)
    assert theme_site.site_vars["head"] == {"head.html"}



import datetime
from datetime import timezone

import pluggy
import pytest
from conftest_feed import feed_test_site
from jinja2 import StrictUndefined

from render_engine.collection import Collection
from render_engine.feeds import RSSFeed
from render_engine.page import Page

pm = pluggy.PluginManager("fake_test")
feed_test_site = feed_test_site


def test_can_manually_set_slug():
    """Test that the feed slug can be set manually"""

    class feed(RSSFeed):
        pages = []
        slug = "test-feed-slug"

    assert feed().slug == "test-feed-slug"


def test_feed_path_name():
    """Test that the feed path name is set correctly"""

    class feed(RSSFeed):
        pages = []

    assert feed().path_name == "feed.rss"


def test_rss_feed_title_from_collection(feed_test_site):
    """
    Test a Collection's title is used as the feed title if no title is provided.

    In this case, the collection is named TestCollection.
    """
    assert "<title>Untitled Site-TestCollection</title>" in feed_test_site


def test_rss_feed_item_url(feed_test_site):
    """Test that the feed item url is set correctly"""
    assert "<link>http://localhost:8000/page.html</link>" in feed_test_site


def test_rss_feed_item_has_guid(feed_test_site):
    """Test that the feed item url is set correctly"""
    assert '<guid isPermaLink="true">http://localhost:8000/page.html</guid>' in feed_test_site


@pytest.mark.skip("Invalid Test")
def test_rss_feed_template_with_strictundefined(engine, tmp_path):
    """Test that the RSS feed template works with the StrictUndefined undefined handler."""
    tmp_dir = tmp_path / "content"
    tmp_dir.mkdir()
    file = tmp_dir / "#"
    file.write_text("test")

    class TestCollection(Collection):
        pages = [Page(content_path=file)]
        Feed = RSSFeed

    collection = TestCollection()
    engine.undefined = StrictUndefined
    rendered_content = collection.feed._render_content(
        engine=engine,
        SITE_TITLE="Test Site Title",
        SITE_URL="http://localhost:8000",
        title="Test Feed Title",
        description="Test Feed Description",
    )
    assert "http://localhost:8000/page.html" in rendered_content


@pytest.mark.parametrize(
    "date",
    [
        datetime.datetime(2023, 4, 15, 0, 0, 0, tzinfo=timezone.utc),
        datetime.datetime(2023, 4, 15, 0, 0, 0, tzinfo=None),
        datetime.date(2024, 4, 15),
    ],
)
def test_rss_feed_template_parses_date_correctly(engine, date):
    """Tests that a feed parses the page date in RFC2822 Format"""

    class TestPage(Page):
        date = datetime.datetime(2023, 4, 15, 0, 0, 0, tzinfo=timezone.utc)  # noqa: UP017

    class TestCollection(Collection):
        Feed = RSSFeed
        pages = [TestPage()]

    collection = TestCollection()

    rendered_content = collection.feed._render_content(
        engine=engine,
        SITE_TITLE="Test Site Title",
        SITE_URL="http://localhost:8000",
        title="Test Feed Title",
        description="Test Feed Description",
    )

    assert "Sat, 15 Apr 2023 00:00:00 +0000" in rendered_content


def test_feed_atom_link(feed_test_site):
    """Test that the feed atom link is set correctly"""
    assert (
        '<atom:link href="http://localhost:8000/testcollection.rss" rel="self" type="application/rss+xml" />'
        in feed_test_site
    )



import pytest

from render_engine import Collection, Page, Site
from render_engine.feeds import RSSFeed


@pytest.fixture()
def site(tmp_path):
    tmp_dir = tmp_path / "content"
    tmp_dir.mkdir()
    file = tmp_dir / "#"
    file.write_text("test")

    site = Site()

    @site.collection
    class TestCollection(Collection):
        pages = [Page(content_path=file)]
        Feed = RSSFeed
        has_archive = True

    return site


@pytest.fixture()
def feed_test_site(tmp_path):
    tmp_dir = tmp_path / "content"
    tmp_dir.mkdir()
    file = tmp_dir / "#"
    file.write_text("test")

    site = Site()

    @site.collection
    class TestCollection(Collection):
        pages = [Page(content_path=file)]
        Feed = RSSFeed
        routes = ["feed"]

    feed = site.route_list["testcollection"].feed
    feed_content = feed._render_content(engine=site.theme_manager.engine, SITE_URL="http://localhost:8000")
    return feed_content



"""Shared Properties and methods across render_engine objects."""

from collections import defaultdict
from collections.abc import Callable

from slugify import slugify


class BaseObject:
    """
    Shared properties for render_engine objects.

    This ensures that the behavior around the title, slug, and path_name are consistent

    This is not intended to be used directly.

    Attributes:
        title (str): The title of the object. If no title is provided, the class name is used.
        template_vars (dict): A dictionary of template variables for the object.
        plugins (list): A list of plugins associated with the object.
        plugin_settings (dict): A dictionary of plugin settings for the object.

    """

    title: str
    template_vars: dict
    plugins: list[Callable] | None
    plugin_settings: dict = {"plugins": defaultdict(dict)}
    skip_site_map: bool = False
    metadata: dict = dict()

    @property
    def _title(self) -> str:
        """
        The title of the object.

        If no title is provided, the class name is used.

        Returns:
            str: The title of the object.
        """
        return getattr(self, "title", self.__class__.__name__)

    @property
    def _slug(self) -> str:
        """
        The slugified path of the object.

        Returns:
            str: The slugified path of the object.

        """
        return slugify(getattr(self, "slug", self._title))

    @staticmethod
    def _metadata_attrs() -> dict[str, str]:
        """attrs used as metadata by the parser"""
        return {"title": "Untitled Entry"}

    @property
    def extension(self) -> str:
        """
        The extension of the object.

        Returns:
            str: The extension of the object.

        """
        return getattr(self, "_extension", ".html")

    @extension.setter
    def extension(self, extension: str) -> None:
        """
        Set the extension of the object.

        Args:
            extension (str): The extension to set.

        """
        self._extension = f".{extension.lstrip('.')}"

    @property
    def path_name(self) -> str:
        """
        Returns the URL path for the object including the extension.

        Returns:
            str: The URL path for the object.

        """
        return f"{self._slug}{self.extension}"

    def url_for(self):
        """
        Placeholder method for generating the URL for the object.

        This method should be implemented in subclasses.

        """
        pass

    def to_dict(self):
        """
        Returns a dictionary of the object's attributes.

        This method is often used to pass attributes into the object's template.

        Returns:
            dict: A dictionary of the object's attributes.

        """
        base_dict = {
            **vars(self),
            "title": self._title,
            "slug": self._slug,
            "url": self.url_for(),
            "path_name": self.path_name,
        }

        # Pull out template_vars
        if hasattr(self, "template_vars"):
            for key, value in self.template_vars.items():
                base_dict[key] = value

        # Pull out plugin_settings
        if hasattr(self, "plugin_settings"):
            for key, value in self.plugin_settings.items():
                base_dict[key] = value

        return base_dict

    def render(self, *args, **kwargs):
        """Render method. Implemented in child objects"""
        raise NotImplementedError(f'{self.__class__.__name__} does not implement the "render" method.')



import pathlib
import textwrap

import pluggy
import pytest
from render_engine_parser import BasePageParser

from render_engine.collection import Collection
from render_engine.page import Page

pm = pluggy.PluginManager("fake_test")


def test_collection_information_parser_passes_to_page(tmp_path):
    """
    Tests that information page parser is passed into the page
    """

    class SimpleBasePageParser(BasePageParser):
        pass

    class BasicCollection(Collection):
        Parser = SimpleBasePageParser
        content_type = Page

    collection = BasicCollection()
    page = collection.get_page()
    assert page.Parser == SimpleBasePageParser


def test_pages_generate_from_collection_content_path(tmp_path: pathlib.Path):
    """
    Tests that pages are generated from a collection
    """

    dir = tmp_path / "content"
    dir.mkdir()

    content = ["foo", "bar", "biz", "baz"]
    for count, val in enumerate(content):
        dir.joinpath(f"test{count}.md").write_text(val)

    class BasicCollection(Collection):
        content_path = dir

    collection = BasicCollection()
    assert len([page for page in collection]) == len(content)

    for page in collection:
        # Order is not guaranteed
        assert page.content in content


def test_collection_archive_no_items_per_page(caplog, tmp_path: pathlib.Path):
    """
    Tests that archive generates a single page if items_per_page is not set
    """

    tmp_dir = tmp_path / "content"
    tmp_dir.mkdir()
    file = tmp_dir / "test.md"
    file.write_text("test")

    class BasicCollection(Collection):
        content_path = tmp_dir.absolute()

    with caplog.at_level("WARNING"):
        collection = BasicCollection()
        assert len(list(collection.archives)) == 1
        assert "`has_archive` is set to `False`" in caplog.text
        assert "BasicCollection" in caplog.text


def test_collection_context(tmp_path: pathlib.Path):
    """
    Tests that `collection` context is passed to the page objects
    """

    tmp_dir = tmp_path / "content"
    tmp_dir.mkdir()
    file = tmp_dir / "test.md"
    file.write_text("test")

    class BasicCollection(Collection):
        content_path = tmp_dir.absolute()
        archive_template = None

    collection = BasicCollection()

    for page in collection:
        assert page.collection["title"] == collection._title


def test_collection_archives_has_title_of_collection(tmp_path: pathlib.Path):
    """
    Tests that the title of the Archive Collection is the same as the parent Collection.

    This should be the case with all archive pages generated

    Issue #105
    """
    tmp_dir = tmp_path / "content"
    tmp_dir.mkdir()

    file1 = tmp_dir / "test.md"
    file1.write_text("test1")

    file2 = tmp_dir / "test2.md"
    file2.write_text("test")

    class BasicCollection(Collection):
        content_path = tmp_dir.absolute()
        items_per_page = 1

    collection = BasicCollection()
    assert len(list(collection.archives)) == 3
    for archive in collection.archives:
        assert archive._title == collection._title


def test_collection_paginated_archives_start_at_1(tmp_path: pathlib.Path):
    """Tests that the first archive page is page 1 and not page 0"""

    tmp_dir = tmp_path / "content"
    tmp_dir.mkdir()

    file1 = tmp_dir / "test.md"
    file1.write_text("test1")

    file2 = tmp_dir / "test2.md"
    file2.write_text("test")


@pytest.mark.parametrize(
    "attr,attrval",
    [
        ("template", "test.html"),
        ("routes", ["/test/long/route"]),
    ],
)
def test_collection_attrs_pass_to_page(tmp_path, attr: str, attrval: str | list[str]):
    """Tests that the template attribute for the collection is passed to the page"""

    class SimpleBasePageParser(BasePageParser):
        pass

    class BasicCollection(Collection):
        PageParser = SimpleBasePageParser
        content_type = Page

    setattr(BasicCollection, attr, attrval)

    collection = BasicCollection()
    page = collection.get_page()

    assert getattr(page, attr) == attrval


def test_collection_sort_by__title():
    """
    Tests that collections are sorted by their `_title`
    """

    class PageA(Page):
        title = "Page A"
        content = "This Page should list first"

    class PageB(Page):
        title = "Page B"
        content = "This Page should list second"

    class PageC(Page):
        title = "Page C"
        content = "This Page should list third"

    # Create instances of the page classes
    page_a = PageA()
    page_b = PageB()
    page_c = PageC()

    # Expected sorted order
    expected_pages = [page_a, page_b, page_c]

    # Create a mixed order of pages to test sorting
    mixed_pages = [page_c, page_a, page_b]

    class CustomCollection(Collection):
        pages = mixed_pages

    custom_collection = CustomCollection()
    sorted_pages = list(custom_collection.sorted_pages)

    # Verify the sorted order matches expected order by checking titles
    assert [page.title for page in sorted_pages] == [page.title for page in expected_pages]


def test_collection_custom_sort_by():
    """
    Tests that the sort_by can be changed
    """

    class PageA(Page):
        title = "Page A"
        content = "This Page should list first"
        custom_sort_content = "z"

    class PageB(Page):
        title = "Page B"
        content = "This Page should list second"
        custom_sort_content = "4"

    class PageC(Page):
        title = "Page C"
        content = "This Page should list third"
        custom_sort_content = "b"

    # Create instances of the page classes
    page_a = PageA()
    page_b = PageB()
    page_c = PageC()

    # Expected order based on custom_sort_content: 4, b, z
    expected_pages = [page_b, page_c, page_a]

    # Create a mixed order of pages
    mixed_pages = [page_a, page_c, page_b]

    class CustomCollection(Collection):
        sort_by = "custom_sort_content"
        pages = mixed_pages

    custom_collection = CustomCollection()
    sorted_pages = list(custom_collection.sorted_pages)

    # Verify the sorted order by checking custom_sort_content values
    assert [page.custom_sort_content for page in sorted_pages] == [page.custom_sort_content for page in expected_pages]


def test_collection_sort_by_error_missing():
    """
    Tests that a custom error message is raised when pages are missing
    the attribute specified in sort_by.
    """

    class PageA(Page):
        title = "Page A"
        content = "This Page is fine"
        custom_sort_content = "z"

    class PageB(Page):
        title = "Page B"
        content = "This Page should error because it is missing the sort_by"

    # Create instances of the page classes
    page_a = PageA()
    page_b = PageB()

    class SortByErrorCollection(Collection):
        sort_by = "custom_sort_content"
        pages = [page_b, page_a]

    custom_collection = SortByErrorCollection()

    with pytest.raises(AttributeError) as excinfo:
        custom_collection.sorted_pages

    # Check that the error message contains helpful information
    assert "Cannot sort pages" in str(excinfo.value)
    assert "custom_sort_content" in str(excinfo.value)
    assert "SortByErrorCollection" in str(excinfo.value)


def test_collection_custom_sort_by_list():
    """
    Tests that the sort_by can be changed
    """

    class PageA(Page):
        title = "Page A"
        content = "This Page should list first"
        custom_sort_content = "z"
        other_sort_content = 1

    class PageB(Page):
        title = "Page B"
        content = "This Page should list second"
        custom_sort_content = "4"
        other_sort_content = 1

    class PageC(Page):
        title = "Page C"
        content = "This Page should list third"
        custom_sort_content = "b"
        other_sort_content = 1

    # Create instances of the page classes
    page_a = PageA()
    page_b = PageB()
    page_c = PageC()

    # Expected order based on custom_sort_content: 4, b, z
    expected_pages = [page_b, page_c, page_a]

    # Create a mixed order of pages
    mixed_pages = [page_a, page_c, page_b]

    class CustomCollection(Collection):
        sort_by = ["other_sort_content", "custom_sort_content"]
        pages = mixed_pages

    custom_collection = CustomCollection()
    sorted_pages = list(custom_collection.sorted_pages)

    # Verify the sorted order by checking custom_sort_content values
    assert [page.custom_sort_content for page in sorted_pages] == [page.custom_sort_content for page in expected_pages]


def test_collection_custom_sort_by_list_with_date():
    """
    Tests that the sort_by can be changed
    """

    class PageA(Page):
        title = "Page A"
        content = "This Page should list first"
        custom_sort_content = "z"
        other_sort_content = 1
        date = "2025-06-01"

    class PageB(Page):
        title = "Page B"
        content = "This Page should list second"
        custom_sort_content = "4"
        date = "2025-06-01"
        other_sort_content = 1

    class PageC(Page):
        title = "Page C"
        content = "This Page should list third"
        custom_sort_content = "b"
        other_sort_content = 1
        date = "2025-06-02"

    # Create instances of the page classes
    page_a = PageA()
    page_b = PageB()
    page_c = PageC()

    # Expected order based on custom_sort_content: 4, b, z
    expected_pages = [page_b, page_a, page_c]

    # Create a mixed order of pages
    mixed_pages = [page_a, page_c, page_b]

    class CustomCollection(Collection):
        sort_by = ["other_sort_content", "date", "custom_sort_content"]
        pages = mixed_pages

    custom_collection = CustomCollection()
    sorted_pages = list(custom_collection.sorted_pages)

    # Verify the sorted order by checking custom_sort_content values
    assert [page.custom_sort_content for page in sorted_pages] == [page.custom_sort_content for page in expected_pages]


@pytest.mark.parametrize(
    "content, context, expected",
    [
        (
            "",
            {"title": "title"},
            textwrap.dedent("""---
title: title
---

Hello, world!"""),
        ),
        (
            "test",
            {},
            textwrap.dedent("""---
title: Untitled Entry
---

test"""),
        ),
        (
            "",
            {},
            textwrap.dedent("""---
title: Untitled Entry
---

Hello, world!"""),
        ),
    ],
)
def test_create_entry(tmp_path: pathlib.Path, content, context, expected):
    """Test the create_entry method"""
    tmp_dir = tmp_path / "content"
    tmp_dir.mkdir()

    class BasicCollection(Collection):
        content_path = tmp_dir.absolute()

    filename = "test.md"
    filepath = tmp_dir / filename
    assert str(filepath) in BasicCollection().create_entry(
        filepath=filepath, editor=None, content=content, metadata=context
    )
    assert filepath.read_text().strip() == expected


def test_create_entry_no_filename():
    """Test create_entry with no filename raises an exception"""
    # Since the base Collection object uses a FileContentManager
    with pytest.raises(ValueError):
        Collection().create_entry()



import json
from collections.abc import Callable
from pathlib import Path
from typing import Any, cast

from render_engine._base_object import BaseObject
from render_engine.plugins import PluginManager


class DataObject(BaseObject):
    """Class to store a data object"""

    data_object: Any = None  # The data object
    serializer: Callable
    serializer_args: dict = {}

    routes: list[str | Path] = ["./"]
    path_name: str | Path = Path("data_object.json")
    plugin_manager: PluginManager | None
    site = None  # This is a Site but circular imports so we can't actually type hint it.

    def __init__(self, serializer: Callable = json.dumps):
        """Ensure that the serializer method is static"""
        self.serializer = staticmethod(serializer)

    @property
    def filename(self) -> str:
        if not isinstance(self.path_name, Path):
            self.path_name = Path(self.path_name)
        return self.path_name.name

    def render(self, *args, **kwargs):
        """
        Renders the data_object to the file at output_file
        """
        from .site import Site

        site: Site = cast(Site, self.site)
        for route in self.routes:
            path = Path(site.output_path, route, self.path_name)
            path.parent.mkdir(parents=True, exist_ok=True)

            settings = dict()
            if (pm := getattr(self, "plugin_manager", None)) and pm is not None:
                settings = {**site.plugin_manager.plugin_settings, "route": self.path_name}
                pm.hook.render_content(page=self, settings=settings, site=self.site)

            data_object = self.data_object
            serialized = (
                self.serializer(data_object, **self.serializer_args)
                if self.serializer_args
                else self.serializer(self.data_object)
            )

            if pm is not None:
                pm.hook.post_render_content(page=self.__class__, settings=settings, site=self.site)

            path.write_text(serialized)



from jinja2.environment import Environment
from jinja2.loaders import ChoiceLoader, DictLoader, FileSystemLoader, PrefixLoader

from render_engine.themes import Theme, ThemeManager


def test_ThemeManager_registers_theme():
    loader1 = DictLoader({"test1.html": "This is a TeSt"})
    loader2 = DictLoader({"test2.html": "This is a {{'TeSt'|test_up}}"})
    loader3 = DictLoader({"test3.html": "This is a {{'TeSt'|test_down}}"})

    loader2theme = Theme(
        prefix="l1",
        loader=loader2,
        static_dir="test",
        filters={"test_up": lambda x: x.upper()},
        plugins=[],
    )

    loader3theme = Theme(
        prefix="l2",
        loader=loader3,
        static_dir="test",
        filters={"test_down": lambda x: x.lower()},
        plugins=[],
    )

    loader4theme = Theme(
        prefix="l3",
        loader=loader1,
        filters={},
        plugins=[],
        template_globals={"head": "index.html"},
    )

    thememgr = ThemeManager(
        engine=Environment(loader=ChoiceLoader([FileSystemLoader("templates"), PrefixLoader({})])),
        output_path="test",
    )
    for x in (loader2theme, loader3theme, loader4theme):
        thememgr.register_theme(x)

    assert loader2theme.prefix in thememgr.prefix
    assert "test3.html" in thememgr.prefix[loader3theme.prefix].list_templates()
    assert isinstance(thememgr.engine.loader, ChoiceLoader)
    assert loader3theme.static_dir in thememgr.static_paths



from pathlib import Path

from render_engine import Collection
from render_engine.content_managers import FileContentManager


def test_find_entry(tmp_path):
    content_path = Path(tmp_path, "test-collection")
    content_path.mkdir(parents=True)
    filepath = content_path / "content.md"
    content = "Test new content"
    filepath.write_text(content)

    class TestCollection(Collection):
        content_path = Path(tmp_path, "test-collection")
        ContentManager = FileContentManager

    collection = TestCollection()
    page = collection.content_manager.find_entry(content_path=filepath)
    assert page is not None
    assert page.content_path == filepath
    assert page.content == content


def test_find_no_valid_entry(tmp_path):
    content_path = Path(tmp_path, "test-collection")
    content_path.mkdir(parents=True)
    filepath = content_path / "content.md"
    content = "Test new content"
    filepath.write_text(content)

    class TestCollection(Collection):
        content_path = Path(tmp_path, "test-collection")
        ContentManager = FileContentManager

    collection = TestCollection()
    page = collection.content_manager.find_entry(content_path=filepath, test_attr="Test")
    assert page is None
    page = collection.content_manager.find_entry(content_path=filepath)
    assert page is not None


def test_find_entry_no_entries(tmp_path):
    content_path = Path(tmp_path, "test-collection")
    content_path.mkdir(parents=True)
    filepath = content_path / "content.md"

    class TestCollection(Collection):
        content_path = Path(tmp_path, "test-collection")
        ContentManager = FileContentManager

    collection = TestCollection()
    page = collection.content_manager.find_entry(content_path=filepath)
    assert page is None


def test_update_entry(tmp_path):
    content_path = Path(tmp_path, "test-collection")
    content_path.mkdir(parents=True)
    filepath = content_path / "content.md"
    content = "Test new content"
    updated = "Test updated content"
    filepath.write_text(content)

    class TestCollection(Collection):
        content_path = Path(tmp_path, "test-collection")
        ContentManager = FileContentManager

    collection = TestCollection()
    assert len(collection.content_manager) == 1
    page = collection.content_manager.find_entry(content_path=filepath)
    assert filepath.read_text() == page.content
    collection.content_manager.update_entry(page, content=updated, test_attr="Test")
    assert filepath.read_text() == f"---\ntest_attr: Test\n---\n\n{updated}"
    page = collection.content_manager.find_entry(content_path=filepath)
    assert page.content == updated
    assert page.test_attr == "Test"
    assert len(collection.content_manager) == 1



from dataclasses import dataclass, field


@dataclass
class Link:
    """
    An opinionated format to reference links. Great for creating links that you will pass into multiple page objects.

    Jinja will pass the string representation of the link to the template.

    Attributes:
        text (str): The text to be displayed for the link.
        url (str, optional): The URL that the link points to. Defaults to "#".
        meta (dict[str, str], optional): Meta variables that translate to attributes on the HTML element.
            Defaults to an empty dictionary.

    Returns:
        str: The string representation of the link.

    Examples:
        link = Link(
            text="Render Engine",
            url="https://render-engine.site",
            meta={"class": "link-class", "id": "custom-link-id"}
        )

        # When rendered in a template, this will be:
        # <a href="https://render-engine.site" class="link-class" id="custom-link-id">Render Engine</a>
    """

    text: str = field(default_factory=str)
    url: str = field(default="#")
    meta: dict[str, str] = field(default_factory=dict)

    def __str__(self) -> str:
        if self.meta:
            attrs = " ".join([f'{k}="{v}"' for k, v in self.meta.items()])
            starting_path = f'<a href="{self.url}" {attrs}>'
        else:
            starting_path = f'<a href="{self.url}">'

        return f"{starting_path}{self.text}</a>"


@dataclass
class Image(Link):
    """
    Link object formatted as an image

    Attributes:
        text (str): The text to be displayed as the image's alternative text.
        url (str): The URL of the image.
        meta (dict, optional): Additional attributes to be added to the image tag.

    Example:
        The following example demonstrates how to create an Image object and render it in a template:

        ```python
        myImage = Link(
            text="Render Engine",
            url="https://render-engine.site",
            meta={"class"="link-class", "id"="custom-link-id"}
        )

        # When rendered in a template

        <img src="https://render-engine.site" alt="Render Engine" id="custom-link-id" class="link-class" />
        ```
    """

    def __str__(self) -> str:
        """Returns the HTML representation of the Image object."""
        if self.meta:
            attrs = " ".join([f'{k}="{v}"' for k, v in self.meta.items()])
            return f'<img src="{self.url}" alt="{self.text}" {attrs} />'
        else:
            return f'<img src="{self.url}" alt="{self.text}" />'



import logging
import re
from pathlib import Path
from typing import Any, cast

from jinja2 import Environment, Template
from render_engine_parser.base_parsers import BasePageParser

from render_engine.themes import ThemeManager

from ._base_object import BaseObject
from .plugins import PluginManager

logger = logging.getLogger("Page")


class BasePage(BaseObject):
    """
    This is the Base Page object.

    It was created to allow for the creation of custom page objects.

    This is not intended to be used directly.

    Attributes:
        slug (str): The slug of the page. Defaults to the `title` slugified.
        content: The content to be rendered by the page.
        parser: The Parser used to parse the page's content. Defaults to `BasePageParser`.
        reference: The attribute to use as the reference for the page in the site's route list.
            Defaults to `slug`.
        extension (str): The file extension for the page. Defaults to ".html".
        routes (list[str] | Path): The list of routes for the page. Defaults to ["./"].
        template (str | Template | None): The template to use for rendering the page.
        site: The Site object that owns the page.
        no_prerender: Flag to not prerender the content
    """

    extension: str = ".html"
    routes: list[str | Path] = ["./"]
    template: str | Template | None
    rendered_content: str | None
    _reference: str = "_slug"
    plugin_manager: PluginManager | None
    site = None  # This is a Site but circular imports so we can't actually type hint it.
    no_prerender: bool = False
    collection: dict | None = None

    @property
    def _content(self) -> Any:
        """returns the content of the page."""
        return getattr(self, "content", None)

    @property
    def _data(self) -> Any:
        """returns the content of the page."""
        return getattr(self, "data", None)

    def url_for(self) -> str:
        """
        Returns the URL for the page including the first route.

        This gets the relative URL for a page.

        !!! note

            Pages don't have access to the `Site` attrs.
            You cannot get an abolute URL from a Page object.
            Use {{SITE_URL}} in your templates to get the absolute URL.


        This is the preferred way to reference a page inside of a template.
        """
        if (route := self.routes[0]) == "./":
            return f"/{self.path_name}"
        else:
            return f"/{route}/{self.path_name}"

    def _render_from_template(self, template: Template, **kwargs) -> str:
        """
        Renders the page from a template.

        If the content looks like a template that

        :param template: Template to render
        :param **kwargs: Data to pass into the template for rendering.
        :return: The rendered page
        """
        template_data = {"data": self._data, "content": self._content}
        if site := getattr(self, "site", None):
            template_data["site_map"] = site.site_map
        if not self.no_prerender and isinstance(self._content, str) and re.search(r"{{.*?site_map.*?}}", self._content):
            # If the content looks like a template, try to render it.
            try:
                content_template = Template(self._content)
            except Exception:
                logger.info(f"Failed to parse {repr(self.path_name)} as a template.", exc_info=True)
            else:
                try:
                    template_data["content"] = content_template.render(
                        **{
                            **self.to_dict(),
                            **template_data,
                            **kwargs,
                        }
                    )
                except Exception:
                    logger.info(f"Failed to pre-render {repr(self.path_name)}.", exc_info=True)

        return template.render(
            **{
                **self.to_dict(),
                **template_data,
                **kwargs,
            },
        )

    def _render_content(self, engine: Environment | None = None, **kwargs) -> str:
        """
        Renders the content of the page.

        If there is not content or a template attribute, an error will be raised.
        """
        engine = getattr(self, "engine", engine)
        template = getattr(self, "template", None)

        # Parsing with a template
        if template and engine:
            template = engine.get_template(template)
            return self._render_from_template(template, **kwargs)

        # Parsing without a template
        try:
            if isinstance(self._content, str):
                return self._content

            else:
                raise ValueError(
                    f"Error rendering {self}. \
                                 The returned content attribute must be a string. Got {type(self._content)}"
                )

        except AttributeError:
            raise AttributeError(
                f"{self} does not have a content attribute. You must either provide a template or content."
            )

    def __str__(self):
        return self._slug

    def __repr__(self) -> str:
        return f"<Page: {self._title}>"

    def render(self, theme_manager: ThemeManager) -> int:
        """Render the page to the file system"""
        rc = 0
        from .site import Site

        site: Site = cast(Site, self.site)

        for route in self.routes:
            path = Path(site.output_path) / Path(route) / Path(self.path_name)
            path.parent.mkdir(parents=True, exist_ok=True)
            settings = dict()
            if (pm := getattr(self, "plugin_manager", None)) and pm is not None:
                settings = {**site.plugin_manager.plugin_settings, "route": route}
                pm.hook.render_content(page=self, settings=settings, site=self.site)
            self.rendered_content = self._render_content(theme_manager.engine)
            # pass the route to the plugin settings
            if pm is not None:
                pm.hook.post_render_content(page=self.__class__, settings=settings, site=self.site)

            rc += path.write_text(self.rendered_content)
        return rc


class Page(BasePage):
    """
    The general BasePage object used to make web pages.

    Pages can be rendered directly from a template or generated from a file.

    !!! note

        Not all attributes are defined by default (those that are marked *optional*) but
        will be checked for in other areas of the code.

    When you create a page, you specify variables passed into rendering template.

    Attributes:
        content_path:
            The path to the file that will be used to generate the Page's `content`.
        extension: The suffix to use for the page. Defaults to `.html`.
        engine:
            If present, the engine to use for rendering the page.

            !!! note
                **This is normally not set and the `Site`'s engine will be used.**

        reference:
            Used to determine how to reference the page in the `Site`'s route_list.
            Defaults to `slug`.
        routes: The routes to use for the page. Defaults to `["./"]`.
        template:
            The template used to render the page.
            If not provided, the `Site`'s `content` will be used.
        Parser:
            The parser to generate the page's `raw_content`.
            Defaults to `BasePageParser`.
        title: The title of the page. Defaults to the class name.

    """

    content: Any
    content_path: Path | str | None
    inherit_plugins: bool
    parser_extras: dict[str, Any] | None
    title: str
    Parser: type[BasePageParser] = BasePageParser

    def __init__(
        self,
        content_path: Path | str | None = None,
        content: Any | None = None,
        Parser: type[BasePageParser] | None = None,
    ) -> None:
        """
        Initializes a new Page object.

        Args:
            content_path (Path, str, optional): The path to the file that will be used to generate the Page's `content`.
            content (Any, optional): The content of the page.
            Parser (type[BasePageParser], optional): The parser to generate the page's `raw_content`.
                Defaults to `BasePageParser`.
        """
        if Parser:
            self.Parser = cast(type[BasePageParser], Parser)

        # Parse Content from the Content Path or the Content
        if content_path := (content_path or getattr(self, "content_path", None)):
            self.metadata, self.content = self.Parser.parse_content_path(str(content_path))

        elif content := (content or getattr(self, "content", None)):
            self.metadata, self.content = self.Parser.parse_content(content)

        else:
            self.metadata = {}
            self.content = None

        # Set the attributes
        for key, val in self.metadata.items():
            setattr(self, key.lower(), val)

        self.content_path = content_path

    @property
    def _content(self) -> Any:
        """
        Returns the parsed content of the page.

        Returns:
            Any: The parsed content of the page.
        """
        return self.Parser.parse(self.content, extras=getattr(self, "parser_extras", {}))



from .blog import Blog
from .collection import Collection
from .data_object import DataObject
from .page import Page
from .site import Site

__all__ = ["Blog", "Collection", "Page", "Site", "DataObject"]



from render_engine_markdown import MarkdownPageParser

from .collection import Collection


class Blog(Collection):
    """
    Custom :py:class:`collection.Collection` class with archiving enabled, sort by `date` by default.

    This class represents a blog collection in the render engine. It inherits from the `Collection` class.

    Attributes:
        BasePageParser (class): The base page parser class to use for parsing blog pages.
            Defaults to `MarkdownPageParser`.
        sort_reverse (bool): Flag indicating whether to sort the blog posts in reverse order. Defaults to `True`.
        sort_by (str): The attribute to sort the blog posts by. Defaults to `"date"`.
        has_archive (bool): Flag indicating whether the blog has an archive. Defaults to `True`.
        Feed (class): The feed class to use for generating RSS feeds. Defaults to `RSSFeed`.

    Methods:
        latest(count: int) -> list[Collection]: Returns the latest blog posts from the collection.

    #TODO:
        - Add Support for JSON Feeds
        - Rename the archive items so they are not private
    """

    BasePageParser = MarkdownPageParser
    sort_by = "date"
    has_archive = True
    sort_reverse = True

    def latest(self, count: int = 1) -> list[Collection]:
        """Get the latest post from the collection."""
        latest_pages = list(
            sorted(
                self.__iter__(),
                key=lambda x: getattr(x, self.sort_by),
                reverse=self.sort_reverse,
            )
        )[0:count]
        return latest_pages



import warnings

from render_engine_markdown import MarkdownPageParser  # noqa: F401

warnings.warn(
    "`render_engine.hookspecs` will be deprecated in version 2024.3.1,  \
        Please use `from render_engine_markdown import MarkdownPageParser` instead.",
    DeprecationWarning,
)



import datetime
from email.utils import format_datetime as fmt_datetime
from typing import cast
from urllib.parse import urljoin

from dateutil.parser import parse
from jinja2 import (
    ChoiceLoader,
    Environment,
    FileSystemLoader,
    PackageLoader,
    PrefixLoader,
    pass_environment,
    select_autoescape,
)

from ._base_object import BaseObject
from .collection import Collection
from .page import BasePage

render_engine_templates_loader = ChoiceLoader(
    [
        # Newly Registered Themes 'templates folder' goes here
        FileSystemLoader("templates"),
        PrefixLoader(
            {
                # "prefix": theme.loader
            }
        ),
        PackageLoader("render_engine", "render_engine_templates"),
    ]
)

engine = Environment(
    loader=render_engine_templates_loader,
    autoescape=select_autoescape(["xml"]),
    lstrip_blocks=True,
    trim_blocks=True,
)


def to_pub_date(value: datetime.datetime | datetime.date) -> str:
    """
    Parse information from the given class object.
    """

    if isinstance(value, datetime.date):
        value = parse(value.isoformat())
    return fmt_datetime(value)


engine.filters["to_pub_date"] = to_pub_date


@pass_environment
def format_datetime(
    env: Environment,
    value: datetime.datetime | datetime.date,
    datetime_format: str | None = None,
) -> str:
    """Parse information from the given class object."""
    format: str
    if datetime_format:
        format = datetime_format
    else:
        format = cast(str, env.globals.get("DATETIME_FORMAT", "%Y-%m-%d"))

    return value.strftime(format)


engine.filters["format_datetime"] = format_datetime


@pass_environment
def to_absolute(env: Environment, url: str) -> str:
    site_url: str = cast(str, env.globals.get("SITE_URL"))
    return str(urljoin(site_url, url))


engine.filters["to_absolute"] = to_absolute


@pass_environment
def feed_url(env: Environment, value: str) -> str:
    """Returns the URL for the collections feed"""
    routes = cast(dict[str, BaseObject], env.globals.get("routes"))

    if routes:
        route = cast(Collection, routes[value])
        return route.feed.url_for()

    else:
        raise ValueError("No Route Found")


engine.filters["feed_url"] = feed_url


@pass_environment
def url_for(env: Environment, value: str, page: int = 0) -> str:
    """Look for the route in the route_list and return the url for the page."""
    routes = cast(dict[str, BaseObject], env.globals.get("routes"))

    if "." in value:
        collection, route = value.split(".", maxsplit=1)

        if collection := routes.get(collection, None):
            collection = cast(Collection, collection)
            for page in collection:
                if getattr(page, page._reference) == route:
                    return page.url_for()

    else:
        route: BaseObject | None
        match route := routes.get(value):
            case Collection():
                return list(route.archives)[page].url_for()
            case BasePage():
                return route.url_for()

    raise ValueError(f"{value} is not a valid route.")


engine.filters["url_for"] = url_for



import subprocess
from collections.abc import Iterable
from pathlib import Path

from more_itertools import flatten

from render_engine.content_managers import ContentManager


class FileContentManager(ContentManager):
    """Content manager for content stored on the file system as individual files"""

    def __init__(
        self,
        content_path: Path | str,
        collection,
        include_suffixes: Iterable[str] = ("*.md", "*.html"),
        **kwargs,
    ):
        self.content_path = content_path
        self.include_suffixes = include_suffixes
        self.collection = collection
        self._pages = None

    def iter_content_path(self):
        """Iterate through in the collection's content path."""
        return flatten([Path(self.content_path).glob(suffix) for suffix in self.include_suffixes])

    @property
    def pages(self) -> Iterable:
        if self._pages is None:
            self._pages = [self.collection.get_page(page) for page in self.iter_content_path()]
        yield from self._pages

    @pages.setter
    def pages(self, value: Iterable):
        self._pages = value

    def create_entry(
        self,
        filepath: Path | None = None,
        editor: str | None = None,
        metadata: dict | None = None,
        content: str | None = None,
        update: bool = False,
    ) -> str:
        """
        Create a new entry for the Collection

        :param filepath: Path object for the new entry
        :param editor: Editor to open to edit the entry.
        :param content: The content for the entry
        :param metadata: Metadata for the new entry
        :param update: Allow overwriting the existing file
        """
        if not filepath:
            raise ValueError("filepath needs to be specified.")

        if not update and filepath.exists():
            raise RuntimeError(f"File at {filepath} exists and update is disabled.")

        parsed_content = self.collection.Parser.create_entry(content=content, **(metadata or {}))
        filepath.write_text(parsed_content)
        if editor:
            subprocess.run([editor, filepath])
        return f"New entry created at {filepath} ."

    def update_entry(self, page, *, content: str | None = None, **kwargs) -> str:
        """
        Update an entry

        :param page: Page object to update
        :param content: Content for the updated page
        :param kwargs: Attributes to be included in the updated page
        :return: String indicating that the page was updated.
        """
        self.create_entry(filepath=page.content_path, metadata=kwargs, content=content, update=True)
        if self._pages:
            self._pages = [
                existing_page for existing_page in self._pages if page.content_path != existing_page.content_path
            ]
            self._pages.append(self.collection.get_page(page.content_path))
        return f"Entry at {page.content_path} updated."



import datetime

import jinja2
import pytest

from render_engine.engine import format_datetime, to_pub_date


@pytest.mark.parametrize(
    "format, override, expected",
    (
        (
            {"DATETIME_FORMAT": "%B %d %Y %H:%M"},
            None,
            "September 27 2023 14:00",
        ),  # Test given value in jinja global
        ({}, None, "27 Sep 2023 14:00 UTC"),  # Test default if no global set
        (
            {"DATETIME_FORMAT": "%B %d %Y %H:%M"},
            "%B %d %Y",
            "September 27 2023",
        ),  # Test override
    ),
)
def test_format_datetime(format, override, expected):
    """
    Tests that the datetime filter works with the following format:

    """
    env = jinja2.Environment()
    env.globals.update(format)
    format_datetime(
        env=env,
        value=datetime.datetime(2023, 9, 27, 14, 0, 0),
        datetime_format=override,
    )


@pytest.mark.parametrize(
    "pubdate",
    (
        (datetime.date(2025, 1, 1)),  # Test given value in jinja global
        (datetime.datetime(2025, 1, 1)),  # Test given value in jinja global
    ),
)
def test_to_pubdate(pubdate):
    """
    Tests that the datetime filter works with the following format:

    """
    assert to_pub_date(pubdate) == "Wed, 01 Jan 2025 00:00:00 -0000"



from pathlib import Path
from typing import Any

from jinja2 import Template

from render_engine.plugins import PluginManager

from .page import BasePage


class Archive(BasePage):
    """
    The Archive is a Page object used by the collection that focuses on presenting the Collection's pages.

    Attributes:
        pages (list[BasePage]): The list of pages to include in the archive.
        title (str): The title of the archive.
        template_vars (dict[str, Any]): The template variables to use for rendering the archive.
        routes (list[str | Path]): The routes for where the archive page should be generated.
        archive_index (int, optional): The index of the page in the series of archive pages. Defaults to 0.
        is_index (bool, optional): Indicates whether the archive is the index page. Defaults to False.
        plugin_manager (PluginManager | None, optional): The plugin manager for the archive. Defaults to None.
        template (str | Template, optional): The template to use for rendering the archive.
            Defaults to "archive.html".

    !!! Warning "Not Directly Used"
        The Archive object is not meant to be used directly.
        It is used by the [Collection][src.render_engine.Collection] object.
        Attributes can be used to customize.
    """

    def __init__(
        self,
        title: str,
        pages: list[BasePage],
        template_vars: dict[str, Any],
        routes: list[str | Path],
        archive_index: int = 0,
        is_index: bool = False,
        plugin_manager: PluginManager | None = None,
        template: str | Template | None = "archive.html",
    ) -> None:
        super().__init__()
        self.slug = title
        self.title = title
        self.archive_index = archive_index
        self.is_index = is_index

        if archive_index:
            self.slug = f"{self._slug}{archive_index}"
        self.pages = pages
        self.plugin_manager = plugin_manager
        self.routes = routes
        self.template = template
        self.template_vars = template_vars



import nox

PYTHON_VERSIONS = ["3.14", "3.13", "3.12", "3.11", "3.10"]


@nox.session(python=PYTHON_VERSIONS, venv_backend="uv")
def test(session: nox.Session) -> None:
    """Run the test suite"""
    session.run_install(
        "uv",
        "sync",
        "--extra",
        "dev",
        "--quiet",
        f"--python={session.virtualenv.location}",
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
        external=True,
    )
    session.run("pytest", "tests")



from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable
from pathlib import Path


class ContentManager(ABC):
    """Base ContentManager abstract class"""

    @property
    @abstractmethod
    def pages(self) -> Iterable:
        """The Page objects managed by the content manager"""
        ...

    @pages.setter
    @abstractmethod
    def pages(self, value: Iterable):
        pass

    def __len__(self):
        return len(list(self.pages))

    def __iter__(self) -> Generator:
        """Iterator for the ContentManager"""
        yield from self.pages

    @abstractmethod
    def create_entry(
        self,
        filepath: Path | None = None,
        editor: str | None = None,
        metadata: dict | None = None,
        content: str | None = None,
    ):
        """Create a new entry"""
        pass

    def find_entry(self, **kwargs):
        """
        Find an entry

        :param kwargs: List of attributes to search by
        :return: Page if it was found otherwise None
        """
        for page in self:
            if all(getattr(page, attr, None) == value for attr, value in kwargs.items()):
                return page
        return None

    @abstractmethod
    def update_entry(self, *, page, **kwargs):
        """Update an entry"""
        pass



import dataclasses
import logging
import pathlib
import shutil
from pathlib import Path
from typing import cast

import slugify
from jinja2 import BaseLoader, ChoiceLoader, Environment, FileSystemLoader


@dataclasses.dataclass
class Theme:
    """
    Base Theme Class for Render Engine

    Attributes:
        loader (BaseLoader): Jinja2 Loader for the theme.
        filters (dict): Dictionary of filters to add to the jinja2 environment.
        prefix (str): Prefix to pass into the prefixLoader.
        plugins (list): List of plugins to add to the site.
        static_dir (str | pathlib.Path | None): Path to static folder.
        template_globals (dict): Dictionary of template globals to add to the jinja2 environment.
            The key is the name of the global and the value is the value of the global.
            In many cases, this will be a string path to a template file
            or a string of template content.

            Example:
                {
                    "head": "head.html",
                    "body_class": "my-class",
                }
    """

    loader: BaseLoader
    prefix: str
    filters: dict = dataclasses.field(default_factory=dict)
    plugins: list = dataclasses.field(default_factory=list)
    template_globals: dict | None = None
    static_dir: str | pathlib.Path | None = None

    def __post_init__(self) -> None:
        if self.prefix:
            self.prefix = slugify.slugify(self.prefix.lower())
        else:
            self.prefix = slugify.slugify(self.__class__.__name__.lower())


@dataclasses.dataclass
class ThemeManager:
    """
    Processes the theme for the site.
    The theme manager is responsible for loading the jinja2 environment and copying static files.

    Attributes:
        engine (Environment): Jinja2 Environment used to render pages.
        output_path (str | Path): Path to write rendered content.
        prefix (dict[str, str]): Dictionary mapping theme prefixes to loader names.
        static_paths (set): Set of filepaths for static folders.
            This will get copied to the output folder. Folders are recursive.
        template_globals (dict[str, set]): Dictionary mapping template global names to sets of values.

    Methods:
        default_template_globals() -> dict[str, set]: Returns the default template globals.
        register_theme(theme: Theme) -> None: Register a theme.
    """

    @staticmethod
    def default_template_globals() -> dict[str, set]:
        return {
            "head": set(),
            "body_class": set(),
        }

    engine: Environment
    output_path: Path | str
    prefix: dict[str, BaseLoader] = dataclasses.field(default_factory=dict)
    static_paths: set = dataclasses.field(default_factory=set)
    template_globals: dict[str, set] = dataclasses.field(default_factory=default_template_globals)

    def register_theme(self, theme: Theme):
        """
        Register a theme.

        Args:
            theme (Theme): Theme object to register.
        """
        logging.info(f"Registering theme: {theme}")
        self.prefix[theme.prefix] = theme.loader

        if theme.static_dir:
            logging.debug(f"Adding static path: {theme.static_dir}")
            self.static_paths.add(theme.static_dir)
        self.engine.filters.update(theme.filters)

        if theme.template_globals:
            for key, value in theme.template_globals.items():
                if isinstance(value, set) and isinstance(self.engine.globals.get(key), set):
                    entry: set = cast(set, self.engine.globals.get(key, set()))
                    entry.update(value)
                    self.engine.globals[key] = entry
                match self.engine.globals.get(key):
                    case set():
                        entry: set = cast(set, self.engine.globals[key])
                        entry.add(value)
                    case _:
                        self.engine.globals[key] = value

    def _render_static(self) -> None:
        """Copies a Static Directory to the output folder"""
        for static_path in self.static_paths:
            logging.debug(f"Copying Static Files from {static_path}")
            if pathlib.Path(static_path).exists():
                shutil.copytree(
                    static_path,
                    pathlib.Path(self.output_path) / pathlib.Path(static_path).name,
                    dirs_exist_ok=True,
                )

    def add_loader(self, idx: int, loader: BaseLoader):
        """Add a loader to the list of loaders"""
        if self.engine.loader is not None and isinstance(self.engine.loader, ChoiceLoader):
            loaders: list[BaseLoader] = cast(list[BaseLoader], self.engine.loader.loaders)
            loaders.insert(idx, loader)

    @property
    def template_path(self) -> str:
        if self.engine.loader is not None and isinstance(self.engine.loader, ChoiceLoader):
            loader: FileSystemLoader = cast(FileSystemLoader, self.engine.loader.loaders[0])
            return loader.searchpath[0]
        return ""



import importlib
import warnings


def test_deperecated_warning():
    with warnings.catch_warnings(record=True) as w:
        importlib.import_module("render_engine.parsers.markdown")
        assert any(isinstance(i.message, DeprecationWarning) for i in w)
